// This file contains cursor management routines shamelessly adapted 
// from Dave Hersey's color printing example from Develop.
// 
// 9/16/94	david	first cut
// 9/20/95	david	improved comments


#include <QuickDraw.h>
#include <Resources.h>
#include <Memory.h>
#include <Events.h>
#include <ToolUtils.h>
//#include <Desk.h>
#include <DiskInit.h>

#include "appGlobals.h"
#include "cursorUtils.h"


/**\
|**| ==============================================================================
|**| PRIVATE TYPEDEFS
|**| ==============================================================================
\**/

/*------------------------------------------------------------------------------*\
	ACurRec
 *------------------------------------------------------------------------------*
		This is just a C structure that parallels the 'acur' animated cursor resource.
		The animated cursor routines use it.
\*------------------------------------------------------------------------------*/
typedef struct ACurRec {
       short	numCursors;		// Number of cursors.
       short	nextFrame;		// Next cursor to display.
       long		cursID[];		// IDs of crsr or CURS resources.
} ACurRec, *ACurPtr, **ACurHdl;


/**\
|**| ==============================================================================
|**| PRIVATE GLOBALS
|**| ==============================================================================
\**/
static ACurHdl			gACur ;					// Spinning cursors' 'acur' resource.
static Handle			gCursor[kMaxNumCursors];// Handles of cursors to spin.
static short			gCursorType[kMaxNumCursors];// "is it a color cursor?" array.
static unsigned long	gTicksLastCursor ;		// TickCount when we last spun cursors.

/**\
|**| ==============================================================================
|**| PUBLIC FUNCTIONS
|**| ==============================================================================
\**/


/*------------------------------------------------------------------------------*\
	SetUpCursors
 *------------------------------------------------------------------------------*
		This routine does initialization for the animated cursor routines
		BumpCursor and TearDownCursors.
		The caller passes in the resource ID of an 'acur' reource, and this routine
		loads the related cursors. We attempt to load 'crsr' resources if available.
		If not, we use black and white 'CURS' resources.
		Finally, it calls BumpCursor to start using the first cursor.
\*------------------------------------------------------------------------------*/
void SetUpCursors ( short acurID )
{
	short	oldRes, cu;
	
	// Save the current resource file, then switch to our application's
	oldRes = CurResFile() ;
//	UseResFile(gAppResRefNum) ;
	
	// get the requested 'acur' resource
	gACur = (ACurHdl) GetResource('acur', acurID) ;

	// If we loaded the resource, detach it and get, load and detach each cursor
	// it points to.  We store the cursors in global handles so that we can easily
	// access them.  If the acur resource references more cursors than we've
	// allocated space for, then we just use as many as we can handle.  Note that
	// we try to load 'crsr' resources first, and if that fails we try 'CURS'
	// resoruces.  That's so we can use color cursors if they're available.
	
	if (gACur != nil)
	{
		DetachResource((Handle) gACur) ;
		
		if( kMaxNumCursors < (*gACur)->numCursors)
			(*gACur)->numCursors = kMaxNumCursors ;

		for ( cu=0; cu < (*gACur)->numCursors; cu++ )
		{
			gCursor[cu] = (Handle) GetCCursor((*gACur)->cursID[cu] >>16) ;
			if ( gCursor[cu] )
				gCursorType[cu] = 2;
			else
			{
				gCursor[cu] = (Handle) GetCursor((*gACur)->cursID[cu] >>16) ;
				// We need to lock the non-color cursors because we have 
				// to dereference their handle for SetCursor.
				if ( gCursor[cu] )
				{
					MoveHHi( gCursor[cu] ) ;
					HLock( gCursor[cu] ) ;
					gCursorType[cu] = 1;
				}
				else
					gCursorType[cu] = 0;
			}
		}

		// Set the "next frame" counter and the "tick count when
		// we last changed cursors" to 0.  That's so that we'll
		// immediately display the first cursor, which we do by
		// calling BumpCursor.  Finally, restore the old
		// resource file.

		(*gACur)->nextFrame = gTicksLastCursor = 0;
		BumpCursor() ;
	}
//	UseResFile(oldRes) ;
}


/*------------------------------------------------------------------------------*\
	BumpCursor
 *------------------------------------------------------------------------------*
		This routine displays the next cursor in our animated cursor list.
\*------------------------------------------------------------------------------*/
void BumpCursor ( void )
{
	register short			curFrame;
	register unsigned long	numTicks;

	// If there's an 'acur' resource loaded and we're in the foreground,
	// see if we should change cursors.  If so, (if the time that's passed
	// since the last change is at least kMinCursorTicks) store the new tick
	// count, grab the next cursor to use and bump the frame counter.

	if (gInBackground)
		return ;
		
	if ( gACur!=nil )
	{
		numTicks = TickCount() ;
		if ((numTicks - gTicksLastCursor) < kMinCursorTicks)
			return;
	
		gTicksLastCursor = numTicks;
		curFrame = (*gACur)->nextFrame;
		(*gACur)->nextFrame = ++(*gACur)->nextFrame % (*gACur)->numCursors;

		// Call the appropriate SetCursor" call, based on whether
		// the next cursor is color or black and white.

		if (gCursor[curFrame] != nil)
		{
			if ( gCursorType[curFrame] == 2 )
				SetCCursor((CCrsrHandle) gCursor[curFrame]) ;
			else if ( gCursorType[curFrame] == 1 )
				SetCursor(*(CursHandle) gCursor[curFrame]) ;
		}
	}
}


/*------------------------------------------------------------------------------*\
	TearDownCursors
 *------------------------------------------------------------------------------*
		This routine undoes what SetUpCursors did.  It disposes of any loaded handles
		and sets the cursor to the cursor finalCursorID. finalCursorID can be
		either kArrowCursor (for the arrow cursor) or a valid 'crsr' or 'CURS'
		resource id.
\*------------------------------------------------------------------------------*/
void TearDownCursors ( short finalCursorID )
{
	short		numCursors, cu;
	Handle		newCursor ;

	// If we have cursors loaded, go through and dispose of them all, and
	// their accompanying 'acur' resource.  Note that these aren't really
	// resources, since we called DetachResource on them earlier.

	if (gACur != nil)
	{
		numCursors = (*gACur)->numCursors;

		for (cu = 0; cu < numCursors; cu++)
		{
			if (gCursor[cu]) 
			{
				if ( gCursorType[cu] == 2 )
					DisposeCCursor( (CCrsrHandle) gCursor[cu] );
				else if ( gCursorType[cu] == 1 )
					DisposeHandle( gCursor[cu]) ;
			}
			gCursor[cu] = nil;
		}
		DisposeHandle( (Handle)gACur ) ;
		gACur = nil;
	}

	// If we didn't get switched into the background, set the cursor to
	// whatever is in finalCursorID.  If it's kArrowCursor, we use the
	// global arrow cursor (which doesn't have an ID).  We look for a
	// 'crsr' first, and if that's not found, try to load a 'CURS'.

	if (gInBackground)
		return ;
		
	if ( finalCursorID==kArrowCursor )
		SetCursor( &(qd.arrow) ) ;
	else
	{
		newCursor = (Handle) GetCCursor(finalCursorID) ;
		if ( newCursor!=nil )
			SetCCursor( (CCrsrHandle)newCursor ) ;
		else
		{
			newCursor = (Handle) GetCursor(finalCursorID) ;
			if (newCursor != nil )
				SetCursor( *(CursHandle)newCursor) ;
		}
	}
}
