/*
	File:		StringUtils.c

	Contains:	Various string utilities (formatting).
	Version:	Color Sync 2.5 or more recent.
	Copyright:	 1997 by Apple Computer, Inc., all rights reserved.

	File Ownership:
		DRI:				john calhoun
		Other Contact:		Steve Swen
		Technology:			ColorSync

	Writers:
		(jkc)	john calhoun

	Change History (most recent first):

		 <3>	 9/16/97	jkc		Added segments.
		 <2>	  9/9/97	jkc		Still in alpha.
		 <1>	  9/9/97	jkc		First checked in.
*/


//===============================================================================
//
//								   StringUtils.c
//
//===============================================================================


#include <TextEdit.h>
#include <Errors.h>
#include <TextUtils.h>
#include <Resources.h>
#include <Script.h>

#include "StringUtils.h"


#pragma segment STRINGUTILS


typedef struct xtnd80
{
	SInt16		exp;
	SInt16		man[4];
} xtnd80;


//===================================================================== Functions
//--------------------------------------------------------------------- StringToString
// This routine copies one pstring into another.
// It assumes that the dest string is long enough to hold the source string.

void StringToString (StringPtr src, StringPtr dest)
{
	BlockMove(src, dest, src[0] + 1);
}

//--------------------------------------------------------------------- OSTypeToString
// This routine converts a 4-byte OSType into a pascal string.
// It assumes that the dest string (which will be 6 chars)
// is long enough to hold the result.

void OSTypeToString (OSType type, StringPtr dest)
{
	dest[0] = 6;
	dest[1] = dest[6] = '\'';
	BlockMove(&type, &dest[2], 4);
}

//--------------------------------------------------------------------- FSSpecToString
// This routine converts the full path of a FSSpec into a pascal string.
// If the appendName parameter is true then the FSSpec's filenam is
// appended to the path.
// It assumes that the dest string (which can be up to 255 chars)
// is long enough to hold the result.

void FSSpecToString (FSSpec spec, StringPtr dest, Boolean appendName)
{
	CInfoPBRec	pb;
	Str255		dirName;
	OSErr		err;
	
	dest[0] = 0;									// initialize full pathname
	
	if (appendName)
		pStrCat(dest, spec.name, 255);

	pb.dirInfo.ioNamePtr = dirName;
	pb.dirInfo.ioVRefNum = spec.vRefNum;			// indicat target volume
	pb.dirInfo.ioDrParID = spec.parID;				// initialize parent directory ID
	pb.dirInfo.ioFDirIndex = -1;					// get info about a directory

	do
	{
		pb.dirInfo.ioDrDirID = pb.dirInfo.ioDrParID;
		err = PBGetCatInfoSync(&pb);				// needs error cheching...
		pStrCat(dirName, "\p:", 255);
		pStrIns(dest, dirName, 255);				// assume fullPathName is a Str255
	}
	while (pb.dirInfo.ioDrDirID != fsRtDirID);
}


//--------------------------------------------------------------------- LongHexToString
// This routine converts a long into a pascal string in hex format.
// It assumes that the dest string (which will be 10 chars)
// is long enough to hold the result.

void LongHexToString (SInt32 num, StringPtr dest)
{
	SInt32		n;
	SInt16		i;
	
	dest[0] = 10;
	dest[1] = '0';
	dest[2] = 'x';
	
	for (i=0; i<8; i++)
	{
		n = (num & 15);
		dest[10-i] = (n<=9) ? ('0'+n) : ('A'+n-10);
		num >>= 4;
	}
}

//--------------------------------------------------------------------- VersionToString
// This routine converts a long into a pascal string according to the 
// convention for encoding version numbers.
// It assumes that the dest string (which may be from 4 to 9 chars)
// is long enough to hold the result.

void VersionToString (SInt32 vers, StringPtr dest)
{
	UInt8		len;
	UInt8		digit;
	
	len = 0;
	digit = (vers & 0xF0000000) >> 28;
	if (digit)
		dest[++len] = '0' + digit;
	
	digit = (vers & 0x0F000000) >> 24;
	dest[++len] = '0' + digit;

	dest[++len] = '.';
	
	digit = (vers & 0x00F00000) >> 20;
	dest[ ++len ] = '0' + digit;
	
	digit = (vers & 0x000F0000) >> 16;
	if (digit)
	{
		dest[++len] = '.';
		dest[++len] = '0' + digit;
	}
	
	digit = (vers & 0x0000FF00) >> 8;
	switch (digit)
	{
		case 0x20:
		dest[ ++len ] = 'd';
		break;
		
		case 0x40:
		dest[ ++len ] = 'a';
		break;
		
		case 0x60:
		dest[ ++len ] = 'b';
		break;
	}
	
	digit = (vers & 0x000000F0) >> 4;
	if (digit)
		dest[++len] = '0' + digit;
	
	digit = (vers & 0x0000000F) >> 0;
	if (digit)
		dest[++len] = '0' + digit;
	
	dest[0] = len;
}

/*------------------------------------------------------------------------------*\
	pStrCat
 *------------------------------------------------------------------------------*
		This routine concatonates the src pstring to the end of the dest pstring.
		It will not make the dst string longer than the size parameter.
\*------------------------------------------------------------------------------*/

void pStrCat (StringPtr dst, StringPtr src, UInt8 size)
{
	if (*dst + *src > size)								// make sure were not too big
		*src = size - *dst;
	BlockMove(src + 1, dst + *dst + 1, *src);			// copy string in
	*dst += *src;										// adjust length byte
}

/*------------------------------------------------------------------------------*\
	pStrIns
 *------------------------------------------------------------------------------*
		This routine inserts the src pstring at the beginning of the dest pstring.
		It will not make the dst string longer than the size parameter.
\*------------------------------------------------------------------------------*/

void pStrIns (StringPtr dst, StringPtr src, UInt8 size)
{
	if (*dst + *src > size)								// make sure were not too big
		*dst = size - *src;
	BlockMove(dst + 1, dst + *src + 1, *dst);			// make room for new string
	BlockMove(src + 1, dst + 1, *src);					// copy new string in
	*dst += *src;										// adjust length byte
}

//--------------------------------------------------------------------- FormatExtd80
// This routine converts an extended80 number into a pascal string using
// a 'FMAT' resource and the ExtendedToString() call so that the resulting
// string will be formatted accrding to the user's setting in the Numbers
// control panel.
// It assumes that the dest string (which can be up to 255 chars)
// is long enough to hold the result.

OSErr FormatExtd80 (extended80 *x, SInt16 formatID, StringPtr dest)
{
	NumberParts			parts;
	NumFormatString		**format;
	Intl0Hndl			itl4;
	SInt32				offset, length;
	FormatStatus		status;
	
				// Just in case!
	dest[0] = 0;
	
				// Get default Number Parts resource/offset/length.
	GetIntlResourceTable(iuCurrentScript, smNumberPartsTable,
							(Handle*)&itl4, &offset, &length);
    if (itl4==nil) return paramErr;
	
				// Retrieve default Number Parts Table.
	parts = *(NumberPartsPtr)(*((Handle)itl4) + offset);
	
				// Retrieve canonical format.
	format = (NumFormatString **)GetResource('FMAT', formatID);
	if (format==nil) return resNotFound;
	
				// Lock it.
	HLock((Handle)format);
	
				// Format it!
	status = ExtendedToString(x, *format, &parts, dest);
	
				// Unlock and release the canonical format.
	HUnlock((Handle)format);
	ReleaseResource((Handle)format); 
	
	return status;
}

//--------------------------------------------------------------------- UnformatExtd80
// This routine converts an extended80 number into a pascal string using
// a 'FMAT' resource and the ExtendedToString() call so that the resulting
// string will be formatted accrding to the user's setting in the Numbers
// control panel.
// It assumes that the dest string (which can be up to 255 chars)
// is long enough to hold the result.

OSErr UnformatExtd80 (extended80 *x, SInt16 formatID, StringPtr str)
{
	NumberParts			parts;
	NumFormatString		**format;
	Intl0Hndl			itl4;
	SInt32				offset, length;
	FormatStatus		status;
	
				// Get default Number Parts resource/offset/length.
	GetIntlResourceTable(iuCurrentScript, iuNumberPartsTable, (Handle*)&itl4, 
			&offset, &length);
    if (itl4 == nil)
    	return paramErr;
	
				// Retrieve default Number Parts Table.
	parts = *(NumberPartsPtr)(*((Handle)itl4) + offset);
	
				// Retrieve canonical format.
	format = (NumFormatString **)GetResource('FMAT', formatID);
	if (format == nil)
		return resNotFound;
	
				// Lock it.
	HLock((Handle)format);
	
				// Format it!
	status = StringToExtended(str, *format, &parts, x);
	
				// Unlock and release the canonical format.
	HUnlock((Handle)format);
	ReleaseResource((Handle)format); 
	
	return status;
}

//--------------------------------------------------------------------- Extd80ToLong

void Extd80ToLong (extended80 *x, SInt32 *num)
{
	UInt32		tempLong;
	SInt16		shiftAmount;
	
	tempLong = *((SInt32 *)(((xtnd80*)x)->man));
	shiftAmount = ((xtnd80*)x)->exp - (0x4000 + 32 - 2);
	
	if (shiftAmount > 0)
		tempLong <<= shiftAmount;
	else if (shiftAmount < 0)
		tempLong >>= -shiftAmount;
	
	*num = tempLong;
}

//--------------------------------------------------------------------- LongToExtd80

void LongToExtd80 (SInt32 *num, extended80 *x)
{
	((xtnd80 *)x)->man[0] = HiWord(*num);
	((xtnd80 *)x)->man[1] = LoWord(*num);
	((xtnd80 *)x)->man[2] = 0;
	((xtnd80 *)x)->man[3] = 0;
	((xtnd80 *)x)->exp = 0x4000 + 32 - 2;		// 16414
}

//--------------------------------------------------------------------- Extd80ToFixed

void Extd80ToFixed (extended80 *x, Fixed *num)
{
	UInt32		tempLong;
	SInt16		shiftAmount;
	
	tempLong = *((Fixed*)((xtnd80*)x)->man);
	shiftAmount = ((xtnd80*)x)->exp - (0x4000 + 16 - 2);
	
	if (shiftAmount > 0)
		tempLong <<= shiftAmount;
	else if (shiftAmount < 0)
		tempLong >>= -shiftAmount;
	
	*num = tempLong;
}

//--------------------------------------------------------------------- FixedToExtd80

void FixedToExtd80 (Fixed *num, extended80 *x)
{
	((xtnd80*)x)->man[0] = HiWord(*num);
	((xtnd80*)x)->man[1] = LoWord(*num);
	((xtnd80*)x)->man[2] = 0;
	((xtnd80*)x)->man[3] = 0;
	((xtnd80*)x)->exp = 0x4000 + 16 - 2;		// 16398
}

//--------------------------------------------------------------------- Extd80ToSmallFract

void Extd80ToSmallFract (extended80 *x, SInt16 *num)
{
	UInt32		tempLong;
	SInt16		shiftAmount;
	
	tempLong = *((SInt16*)((xtnd80*)x)->man);
	shiftAmount = ((xtnd80*)x)->exp - (0x4000 + 1 - 2);
	
	if (shiftAmount > 0)
		tempLong <<= shiftAmount;
	else if (shiftAmount < 0)
		tempLong >>= -shiftAmount;
	
	*num = tempLong;
}

//--------------------------------------------------------------------- SmallFractToExtd80

void SmallFractToExtd80 (SInt16 *num, extended80 *x)
{
	((xtnd80*)x)->man[0] = *num;
	((xtnd80*)x)->man[1] = 0;
	((xtnd80*)x)->man[2] = 0;
	((xtnd80*)x)->man[3] = 0;
	((xtnd80*)x)->exp = 0x4000 + 1 - 2;		// 16384
}

//--------------------------------------------------------------------- FormatLong
// This routine converts a long integer into a pascal string using
// a 'FMAT' resource and the ExtendedToString() call so that the resulting
// string will be formatted accrding to the user's setting in the Numbers
// control panel.  It does this by converting the long into an extended80
// and then calling FormatExtd80.
// It assumes that the dest string (which can be up to 255 chars)
// is long enough to hold the result.

OSErr FormatLong (SInt32 num, SInt16 formatID, StringPtr dest)
{
	extended80	x80;
	LongToExtd80(&num, &x80);
	return FormatExtd80(&x80, formatID, dest);
}

//--------------------------------------------------------------------- UnformatLong
// This routine converts a pascal string into  a Long number using
// a 'FMAT' resource and the StringToExtended() call so that the resulting
// string will be unformatted accrding to the user's setting in the Numbers
// control panel.  It does this by calling UnformatExtd80 and then converting 
// the extended80 into a Long.

OSErr UnformatLong (StringPtr str, SInt16 formatID, SInt32 *num)
{
	OSErr		theErr;
	extended80	x80;
	
	theErr = UnformatExtd80(&x80, formatID, str);
	Extd80ToLong(&x80, num);
	
	return theErr;
}

//--------------------------------------------------------------------- FormatFixed
// This routine converts a Fixed number into a pascal string using
// a 'FMAT' resource and the ExtendedToString() call so that the resulting
// string will be formatted accrding to the user's setting in the Numbers
// control panel.  It does this by converting the Fixed into an extended80
// and then calling FormatExtd80.
// It assumes that the dest string (which can be up to 255 chars)
// is long enough to hold the result.

OSErr FormatFixed (Fixed num, SInt16 formatID, StringPtr dest)
{
	extended80	x80;
	
	FixedToExtd80(&num, &x80);
	return FormatExtd80(&x80, formatID, dest);
}

//--------------------------------------------------------------------- UnformatFixed
// This routine converts a pascal string into  a Fixed number using
// a 'FMAT' resource and the StringToExtended() call so that the resulting
// string will be unformatted accrding to the user's setting in the Numbers
// control panel.  It does this by calling UnformatExtd80 and then converting 
// the extended80 into a Fixed.

OSErr UnformatFixed (StringPtr str, SInt16 formatID, Fixed *num)
{
	OSErr		theErr;
	extended80	x80;
	
	theErr = UnformatExtd80(&x80, formatID, str);
	Extd80ToFixed(&x80, num);
	
	return theErr;
}

//--------------------------------------------------------------------- FormatSmallFract
// This routine converts a SmallFract number into a pascal string using
// a 'FMAT' resource and the ExtendedToString() call so that the resulting
// string will be formatted accrding to the user's setting in the Numbers
// control panel.  It does this by converting the SmallFract into an extended80
// and then calling FormatExtd80.
// It assumes that the dest string (which can be up to 255 chars)
// is long enough to hold the result.

OSErr FormatSmallFract (SInt16 num, SInt16 formatID, StringPtr dest)
{
	extended80	x80;
	
	SmallFractToExtd80 (&num, &x80);
	return FormatExtd80(&x80, formatID, dest);
}

//--------------------------------------------------------------------- FormatFloat
// This routine converts a Float number into a pascal string using
// a 'FMAT' resource and the ExtendedToString() call so that the resulting
// string will be formatted accrding to the user's setting in the Numbers
// control panel.  It does this by converting the Float into a Fixed
// and then calling FormatFixed.
// It assumes that the dest string (which can be up to 255 chars)
// is long enough to hold the result.

OSErr FormatFloat (float num, SInt16 formatID, StringPtr dest)
{
	Fixed		theFixed;
	
	theFixed = (SInt32)(num * 65536.0);
	return FormatFixed(theFixed, formatID, dest);
}

//--------------------------------------------------------------------- UnformatFloat
// This routine converts a pascal string into a Float number using
// a 'FMAT' resource and the StringToExtended() call so that the resulting
// string will be unformatted accrding to the user's setting in the Numbers
// control panel.  It does this by calling UnformatFloat and then converting 
// the Fixed into a float.

OSErr UnformatFloat (StringPtr dest, SInt16 formatID, float *num)
{
	Fixed		theFixed;
	OSErr		theErr;
	
	theErr = UnformatFixed(dest, formatID, &theFixed);
	*num = ((float)(theFixed)) / 65536.0;
	
	return theErr;
}

