
//===============================================================================
//
//								WindowCalibratorMain.c
//
//===============================================================================


#include <CMApplication.h>
#include <Resources.h>
#include <Sound.h>

#include "WindowCalibratorEvents.h"
#include "WindowCalibratorMain.h"
#include "WindowCalibratorUtils.h"


pascal Boolean CanCalibrate (AVIDType);
pascal OSErr Calibrate (CalibratorInfo *);
static pascal OSErr ProfileIterate (CMProfileIterateData *, void *);
static OSErr OpenMainWindow (CalibratorGlobals *);
static OSErr CreateNewProfileWithGammaTag (CalibratorGlobals *);
static OSErr CalibrateCommon (CalibratorGlobals *);
void main (void);


//===================================================================== Functions
//--------------------------------------------------------------------- CanCalibrate

pascal Boolean CanCalibrate (AVIDType displayID)
{
#pragma unused (displayID)
	Boolean		weCanCalibrateThisMonitor;
	
	weCanCalibrateThisMonitor = true;
	
				// Here you should determine if your calibrator can work
				// on the display indicated by the AVIDType passed in.
				// There are a host of Display Mananager calls that allow
				// you to get all sorts of information about a monitor
				// requiring only the AVIDType.
				
				// If your calibrator requires additional hardware or additional
				// code libraries - you ought to check for these at this time
				// and return false if you cannot operate.
				
				// Do not put up any alerts or dialogs here!  No UI!
				// If you must present a message to the user it's better that 
				// you "lie" here and return true - you can then put up your
				// alert if the user selects your calibrator from the list.
	
	return weCanCalibrateThisMonitor;
}

//--------------------------------------------------------------------- Calibrate

pascal OSErr Calibrate (CalibratorInfo *theInfo)
{
	CalibratorGlobals	theGlobals;
	OSErr				theErr;
	
				// Grab the display ID and update proc.
	theGlobals.displayID = theInfo->displayID;
	theGlobals.eventProc = theInfo->eventProc;
	
				// Call our common calibration function.
				// This allows us to build this code with a 'main' function
				// and as a stand-alone app.  (We can't do both at once - 
				// have a calibrator plug-in that is also a stand-alone app 
				// because the calibrator must be of type 'shlb' (not 
				// launchable).  Nonetheless, for testing purposes, we can 
				// have a seperate CodeWarrior project that makes the app 
				// out of this same source code.
	
	theErr = CalibrateCommon(&theGlobals);
	
				// The two fields of the CalibratorInfo struct that a 
				// calibrator must fill in are the 'isGood' field and 
				// the 'profileLocation'.  If isGood == true, the profile
				// returned will be made the default profile for the current 
				// monitor.  The profileLocation simply points to the 
				// profile.
				
				// It is advised that you ensure the profileLocation is:
				//    a: a file-based profile (not handle based or other)
				//    b: located somewhere within the ColorSync Profiles 
				//       folder.
				// Otherwise, the newly created profile may fail to show
				// up in the Monitors and Sound ColorSync Profile list.
	
				// Flag the profile as good (the user clicked "Okay").
	theInfo->isGood = theGlobals.createIt;
	if (theInfo->isGood)
		*theInfo->profileLocationPtr = theGlobals.profileLocation;
	
	return theErr;
}

//--------------------------------------------------------------------- ProfileIterate
// This is the function called iteratively by ColorSync.  Each time it 
// is called, it is passed a profile found in the user's ColorSync 
// profiles folder.  We want only RGB profiles intended for a display 
// (monitor).  Since the profile's header is one of the fields passed 
// to us in the CMProfileIterateData structs, we can check the header's 
// 'dataColorSpace' and 'profileClass' fields.

static pascal OSErr ProfileIterate (CMProfileIterateData *profData, void *refCon)
{
	CalibratorGlobals	*theGlobals;
	UInt32				index;
	OSErr				theErr;
	
				// Assume no errors.
	theErr = noErr;
	
				// Set up globals
	theGlobals = (CalibratorGlobals *)refCon;
	
				// Is it an RGB profile for monitors?
	if ((profData->header.dataColorSpace == cmRGBData) &&
			(profData->header.profileClass == cmDisplayClass))
	{
				// If 'allProfileLocations' has been allocated (not nil) 
				// then we're going to add the profile's location to an 
				// array.
		if (theGlobals->allProfileLocations)
		{
			if (theGlobals->count < theGlobals->numProfiles)
				theGlobals->allProfileLocations[theGlobals->count] = profData->location;
			else
				SysBeep(10);
			
			theGlobals->count = theGlobals->count + 1;
		}
		else
		{
				// On the first pass through the iterator, we just want the 
				// the profile's name and a count.
			index = AppendMenuNoMeta(theGlobals->profileMenu, profData->name);
			theGlobals->numProfiles = theGlobals->numProfiles + 1;
		}
	}
	
bail:
	
	return theErr;
}

//--------------------------------------------------------------------- OpenMainWindow
// Bring up a window for our calibrator.

static OSErr OpenMainWindow (CalibratorGlobals *theGlobals)
{
	Rect		bounds;
	
	theGlobals->mainWindow = GetNewWindow(128, 0L, (GrafPtr)(-1L));
	if (!theGlobals->mainWindow)
		return -1;
	
	SetPort(theGlobals->mainWindow);
	
	GetGlobalWindowRect(theGlobals->mainWindow, &bounds);
	CenterRectInRect(&bounds, &(theGlobals->displayBounds));
	MoveWindow(theGlobals->mainWindow, bounds.left, bounds.top, true);
	ShowWindow(theGlobals->mainWindow);
	
				// Stuff the globals into the windows reference constant.
	SetWRefCon(theGlobals->mainWindow, (SInt32)theGlobals);
	
	theGlobals->popUpControl = GetNewControl(128, theGlobals->mainWindow);
	theGlobals->sliderControl = GetNewControl(129, theGlobals->mainWindow);
	theGlobals->okayButton = GetNewControl(130, theGlobals->mainWindow);
	theGlobals->cancelButton = GetNewControl(131, theGlobals->mainWindow);
	
	return noErr;
}

//--------------------------------------------------------------------- CreateNewProfileWithGammaTag
// Here we generate the new profile.  Because this is a rather lame 
// sample app, we don't really carefully craft a new profile that 
// describes this monitor.  Rather, we make a copy of the profile 
// selected in the pop-up, stuff our new video-card gamma tag in, 
// and change the name to "Sample Profile".

static OSErr CreateNewProfileWithGammaTag (CalibratorGlobals *theGlobals)
{
	CMVideoCardGammaType	gammaFormula;
	CMVideoCardGammaFormula	*formulaPtr;
	CMProfileRef			srcProf;
	CMError					theErr;
	UInt32					tagSize;
	Boolean					foundIt;
	
				// Open the profile selected from list and make a copy of it.
	theErr = CMOpenProfile(&srcProf, &theGlobals->profileLocation);
	if (theErr)
		goto skipOut;
	
				// Stuff the name 'Sample Profile' into the file name.
	PasStringCopy("\pSample Profile", theGlobals->profileLocation.u.fileLoc.spec.name);
	
				// If a file with that name already exists - delete it.
	(void *) FSpDelete(&(theGlobals->profileLocation.u.fileLoc.spec));
	
				// Now make a copy of the profile (using the new name 'Sample Profile').
	theErr = CMCopyProfile(&theGlobals->profileRef, &theGlobals->profileLocation, srcProf);
	if (theErr)
		goto skipOut;
	theErr = CMCloseProfile(srcProf);
	if (theErr)
		goto skipOut;
	
				// Change the internal name to 'Sample Profile' as well.
	theErr = SetScriptProfileDescription(theGlobals->profileRef, "\pSample Profile", 0);
	
				// Delete gamma tag if it exists.
	theErr = CMProfileElementExists(theGlobals->profileRef, cmVideoCardGammaTag, &foundIt);
	if (foundIt)
	{
		theErr = CMRemoveProfileElement(theGlobals->profileRef, cmVideoCardGammaTag);
		if (theErr)
			goto skipOut;
	}
	
				// Add gamma table tag.
				// Fill out 'CMVideoCardGammaType' fields.
	gammaFormula.typeDescriptor = cmSigVideoCardGammaType;
	gammaFormula.reserved = 0L;
	gammaFormula.gamma.tagType = cmVideoCardGammaFormulaType;
	formulaPtr = &(gammaFormula.gamma.u.formula);
	
	formulaPtr->redGamma = (Fixed)(theGlobals->theNewGamma.red * 65536.0);
	formulaPtr->redMin = (Fixed)(0.0 * 65536.0);
	formulaPtr->redMax = (Fixed)(1.0 * 65536.0);
	
	formulaPtr->greenGamma = formulaPtr->redGamma;
	formulaPtr->greenMin = formulaPtr->redMin;
	formulaPtr->greenMax = formulaPtr->redMax;
	
	formulaPtr->blueGamma = formulaPtr->redGamma;
	formulaPtr->blueMin = formulaPtr->redMin;
	formulaPtr->blueMax = formulaPtr->redMax;
	
				// Add the new tag.
	tagSize = sizeof(CMVideoCardGammaType);
	theErr = CMSetProfileElement(theGlobals->profileRef, cmVideoCardGammaTag, tagSize, &gammaFormula);
	if (theErr)
		goto skipOut;
	
				// Save off changes to the profile and close it.
	theErr = CMUpdateProfile(theGlobals->profileRef);
	if (theErr)
		goto skipOut;
	theErr = CMCloseProfile(theGlobals->profileRef);
	if (theErr)
		goto skipOut;
	
skipOut:
	
	return theErr;
}

//--------------------------------------------------------------------- CalibrateCommon

static OSErr CalibrateCommon (CalibratorGlobals *theGlobals)
{
	CMProfileIterateUPP	buildMenuProc;
	GDHandle			theDisplay;
	OSErr				theErr;
	UInt8				wasState;
	
				// Determine our screen bounds and pixel depth.
	theErr = DMGetGDeviceByDisplayID(theGlobals->displayID, &theDisplay, false);
	if (theErr != noErr)
		goto bail;
	
	wasState = HGetState((Handle)theDisplay);
	HLock((Handle)theDisplay);
	theGlobals->displayBounds = (*((*theDisplay)->gdPMap))->bounds;
	HSetState((Handle)theDisplay, wasState);
	
				// Bring up the main window and init associated vars.
	theErr = OpenMainWindow(theGlobals);
	if (theErr != noErr)
		goto bail;
	
				// Build the pop-up menu.
	theGlobals->numProfiles = 0;
	theGlobals->count = 0;
	theGlobals->profileMenu = GetMenu(128);
	theGlobals->allProfileLocations = 0L;
	buildMenuProc = NewCMProfileIterateProc(ProfileIterate);
	theErr = CMIterateColorSyncFolder(buildMenuProc, 0L, 0L, (void *)theGlobals);
	
				// Allocate space for all profile locations.
				// Checking for memory allocation errors is particularly important
				// since you are just a plug-in piece living in the Monitor's and 
				// Sound heap space.  If you need more than 500K or so, you ought 
				// to really think about using temp mem.
	theGlobals->allProfileLocations = 
			(CMProfileLocation *)NewPtr(sizeof(CMProfileLocation) * theGlobals->numProfiles);
	if (!theGlobals->allProfileLocations)
	{
		theErr = MemError();
		goto bail;
	}
	
				// Call iterator again - this time saving off locations.
	theErr = CMIterateColorSyncFolder(buildMenuProc, 0L, 0L, (void *)theGlobals);
	if (theErr != noErr)
		goto bail;
	
				// Start out with a final gamma of 1.8 
				// If we assume a gamma of 2.2 for the monitor we will then
				// need a correction of 0.82 on the video card (1.8 / 2.2).
	theGlobals->theNewGamma.red = 0.82;
	theGlobals->theNewGamma.green = 0.82;
	theGlobals->theNewGamma.blue = 0.82;
	theErr = VPTProgramGammaExponent(theGlobals->displayID, &(theGlobals->theNewGamma));
	
				// Set some default values.
	theGlobals->profileLocation = theGlobals->allProfileLocations[0];
	
				// Set the pop-up.
	SetControlMinimum(theGlobals->popUpControl, 1);
	SetControlMaximum(theGlobals->popUpControl, theGlobals->numProfiles);
	SetControlValue(theGlobals->popUpControl, 1);
	
	theGlobals->quitting = false;
	
				// Fall into event loop until user chooses to leave.
	do
	{
		MainEventLoop(theGlobals);
	}
	while (!theGlobals->quitting);
	
				// Did the user click the Create It button?
	if (theGlobals->createIt)
	{
		theErr = CreateNewProfileWithGammaTag(theGlobals);
	}
	else
	{
				// Restore gamma to 1.8.
		theGlobals->theNewGamma.red = 0.82;
		theGlobals->theNewGamma.green = 0.82;
		theGlobals->theNewGamma.blue = 0.82;
		theErr = VPTProgramGammaExponent(theGlobals->displayID, &(theGlobals->theNewGamma));
	}
	
				// Dispose of array of profiles.
	if (theGlobals->allProfileLocations)
		DisposePtr((Ptr)theGlobals->allProfileLocations);
	
				// Clean up remaining structures.
	DisposeWindow(theGlobals->mainWindow);
	
bail:
	
	return theErr;
}

//--------------------------------------------------------------------- main
// This is for building the code as a stand-alone app.  We have to 
// fill in the AVID and eventProc ourselves though.  The eventProc is 
// nil of course since no external app is calling us (i.e. Monitors and 
// Sound).  Out of laziness, I'll just get the AVID of the main monitor.

void main (void)
{
	CalibratorGlobals	*theGlobals;
	GDHandle			theDevice;
	OSErr				theErr;
	
	theGlobals = 0L;
	InitToolbox();
	
				// Allocate our global storage.
	theGlobals = (CalibratorGlobals *)NewPtr(sizeof(CalibratorGlobals));
	if (!theGlobals)
		goto bail;
	
				// Get the display device.
	theDevice = GetMainDevice();
	theErr = DMGetDisplayIDByGDevice(theDevice, &(theGlobals->displayID), true);
	if (theErr)
		theGlobals->displayID = 0;
	
	theGlobals->eventProc = 0L;
	
				// Call the common calibration routine.
	(void) CalibrateCommon(theGlobals);
	
bail:
	
				// Clean up our own allocation.
	if (theGlobals)
		DisposePtr((Ptr)theGlobals);
}

