/*
	File:		FileUtils.c
	Contains:	File access and copy Routines ProfileAccess engine use
	Copyright:	 1995-1998 by Apple Computer, Inc., all rights reserved.
	Version:	ColorSync 2 or later

	File Ownership:
		DRI:				David Hayward
		Other Contact:		Steve Swen
		Technology:			ColorSync 2

	Writers:

		(ebb)	Eric Broadbent		(jkc)	john calhoun
		(DH)	David Hayward

	Change History (most recent first):

		<23>	 5/13/98	ebb		Adding ProfLocationFromFileSpec().
	   <22*>	  5/7/98	ebb		Make FileSpecCopy nil-safe, fix compile error.
	   								Avoid name collision, rename IsAliasFile to MyIsAliasFile.
	    <22>	  5/8/98	WN		New header names.
	    <21>	  5/6/98	WN		updated TARGET_OS_WIN32 support.
		<20>	  5/5/98	ebb		Get/SetEOF abstractions.
		<19>	 3/13/98	ebb		Use Smart memory routines.
									Abstracted some file system calls.
									Collapsed FSpGetFileLocation and GetFileLocation.
									ResolveFSSpec -> ResolveFileSpec
									NewAliasFile, IsAliasFile are Mac-only.
									CopyFork -> CopyFile
									FolderCreate not needed on unix?
		<18>	  2/5/98	DH		Cleaned-up resource chain set/restore.
		<17>	 1/30/98	DH		Added InfoCopy (used by scripting plugins).
		<16>	12/23/97	DH		Added RsrcCopy routine.
		<15>	12/17/97	DH		Fix resource fork bug
		<14>	11/19/97	jkc		Resource chain was being screwed up by
									ResolveAliasFileMountOption().
		<13>	 9/25/97	DH		Omit needless variables.
		<12>	 9/19/97	DH		Added some new handy routines.
		<11>	 9/15/97	DH		oops.
		<10>	 9/15/97	DH		Added FileClose()
		 <9>	 7/21/97	DH		Added CreateTempFile.
		 <8>	 6/11/97	jkc		Added debug asserts to GetResourceCSResFile() since I'm dying
									here and want to know why.
		 <7>	 7/10/96	DH		Changed ResolveAliasFile to ResolveAliasFileMountOption. (Radar
									#1366007)
		 <6>	 6/11/96	DH		Fixed bug in CopyFromToChunk() when srcSize != dstSize.
		 <5>	  6/7/96	DH		Added ReadPos and WritePos
		 <4>	 2/24/96	DH		Changed GetMaxPtr to SmartNewPtrMax
		 <3>	 2/12/96	DH		Added GetFileLocation and FSpGetFileLocation
		--->  Older change history at bottom of file  <--- 
	To do:
		Remove Refrence to ColorSync Creator and Type

*/


// General Headers
#include <MacTypes.h>
#include <Errors.h>
#include "DebugAssert.h"
#include <Files.h>
#include <Devices.h>
#include <Events.h>
#include <Aliases.h>
#include <OSUtils.h>
#include <Resources.h>
#include <Folders.h>
#include <TextUtils.h>
#include <Script.h>

// General ColorSync includes
#include <CMApplication.h>
#include "FileUtils.h"
#include "MemUtils.h"



// Fixme - Remove these two lines when they are added to CMApplication.h
// ebb 04/25/98
#ifndef	kColorSyncCreator
#define kColorSyncCreator			'sync'
#endif
#ifndef	kColorSyncFileType
#define kColorSyncFileType			'prof'
#endif
#ifndef	kProfileFileType
#define	kProfileFileType			kColorSyncFileType
#endif

#define kDefaultFileBufferSize		512
#define kMinimumBufferSize			16
#define kNoCacheBit					32

static OSErr MyIsAliasFile (const FileSpec *, Boolean *, Boolean *);

//=====================================================================
//	Static Function Prototypes
//=====================================================================
static OSErr CopyFile			(			FileRef		srcRefNum,
											FileRef		dstRefNum,
											void		*copyBufferPtr,
											SInt32		copyBufferSize );
						
//===================================================================== Generic
//===================================================================== Functions

#pragma mark <<<<<< Public Functions >>>>>>

//--------------------------------------------------------------------- ResolveFSSpec
//	Abstract: 
//		This routine is called internally to resolve aliases before
//		operating on files.
//
//	Input:
//		*fromSpec		the original fileSpec
//
//	Output:
//		*toSpec			set to the final, de-aliased file.
//---------------------------------------------------------------------

OSErr ResolveFileSpec (FileSpec *fromSpec, FileSpec *toSpec)
{
	OSErr		theErr = noErr;
	Boolean		isFolder;
	Boolean		wasAlias;
	
	theErr = FSMakeFSSpec(fromSpec->vRefNum,fromSpec->parID, (StringPtr) &fromSpec->name,toSpec);
	if (!theErr)
		theErr = ResolveAliasFileMountOption( toSpec, true, &isFolder, &wasAlias, false);
	else
		theErr = noErr;
	
	return theErr;
}


//--------------------------------------------------------------------- RsrcOpenAndGetEOF
//	Abstract: 
//		Multi-purpose resource fork opening routine which 
//		will, optionally return the file length.
//
//	Input:
//		*spec					the filespec for the file.  We will resolve 
//								any aliases, if necessary.
//		perm					read/write permission to open the file with.
//
//	Output:
//		*refNum					file reference number for the file, if we opened.
//		*length					if not nil, returns the length of the file opened.
//
//	Return:
//		noErr, or any file error.
//---------------------------------------------------------------------

OSErr RsrcOpenAndGetEOF (	const	FileSpec		*spec,
									short			perm,
									FileRef			*refNum,
									long			*length )
{
	OSErr		theErr;
	FileSpec	realSpec;
	
	*refNum = 0;
	if (length) *length = 0;
	
	theErr = ResolveFileSpec ((FileSpec*) spec, &realSpec);
	if (!theErr)
	{
		theErr = FSpOpenRF( &realSpec, perm, refNum);
		if (!theErr && length )
		{
			theErr = FileGetEOF(*refNum,(unsigned long*)length);
			if (theErr)
				(void) FileClose(*refNum);
		}
	}
	return theErr;
}

//--------------------------------------------------------------------- RsrcCopy
//	Abstract:
//		Replace the contents of the destination resource fork with
//		that of the source resource fork.  We try to allocate a buffer
//		equal to the size of the entire source file.
//		
//	Params:
//		srcSpec		(in)	Source file
//		dstSpec		(in)	Destination file
//		
//	Return:
//		noErr					successful
//		FileMgr Error
//		MemFullErrr
//---------------------------------------------------------------------

OSErr RsrcCopy (const FileSpec *srcSpec, const FileSpec *dstSpec)
{
	SInt16	srcRef=0, dstRef=0;
	SInt16	theErr;
	SInt32	maxSize;
	Ptr		copyBufferPtr;
	SInt32	buffSize;
	
	theErr = RsrcOpenAndGetEOF( srcSpec, fsCurPerm, &srcRef, &maxSize );
	if (!theErr)
		theErr = RsrcOpenAndGetEOF( dstSpec, fsRdWrPerm, &dstRef, nil );
	
	if (!theErr)
	{
		copyBufferPtr = SmartNewPtrMax( maxSize,&theErr, &buffSize);
		if (copyBufferPtr)
			theErr = CopyFile(srcRef,dstRef,copyBufferPtr,buffSize);
		else
			theErr = SmartMemError(copyBufferPtr);
		SmartDisposePtr( &copyBufferPtr );
	}
	
	(void) FileClose( srcRef );
	(void) FileClose( dstRef );
		
	return theErr;
}

//--------------------------------------------------------------------- InfoCopy
//	Abstract:
//		Copy the Finder's file attributes from the source to the dest
//		
//	Params:
//		srcSpec		(in)	Source file
//		dstSpec		(in)	Destination file
//		
//	Return:
//		noErr					successful
//		FileMgr Error
//		MemFullErrr
//---------------------------------------------------------------------
OSErr InfoCopy (const FileSpec *srcSpec, const FileSpec *dstSpec)
{
	OSErr		theErr;
	FInfo		srcInfo;
	FInfo		dstInfo;
	UInt16		attributes;
	
	theErr = FSpGetFInfo( srcSpec, &srcInfo);
	require( theErr==noErr, bail);

	theErr = FSpGetFInfo( dstSpec, &dstInfo);
	require( theErr==noErr, bail);
	
	attributes = kColor | kHasCustomIcon | kIsStationery | kHasBundle;
	
	dstInfo.fdFlags &= ~attributes;
	dstInfo.fdFlags |= (srcInfo.fdFlags & attributes);

	theErr = FSpSetFInfo( dstSpec, &dstInfo);
	require( theErr==noErr, bail);
	
bail:
	
	return theErr;
}

//--------------------------------------------------------------------- ResolveAliasFileMountOption
//	Resolves a Finder alias file, with the option of not mounting remote 
//	volumes in the process.
//	See Tech Note 315.
//	
//	spec					on entry, the location of the Finder alias file
//							on exit, the location of the real file or folder
//	resolveAliasChains		whether to resolve all the way down to the real file
//	targetIsFolder			on exit, indicates whether the alias target is a folder
//	wasAliased				on exit, indicates whether the initial file was an alias
//	mountRemoteVols			indicates whether to mount remote volumes while resolving
//	
//	returns					noErr if successful, or an error code
//---------------------------------------------------------------------

OSErr ResolveAliasFileMountOption	(	FileSpec	*spec,
										Boolean		resolveAliasChains,
										Boolean		*targetIsFolder,
										Boolean		*wasAliased,
										Boolean		mountRemoteVols )
{
				// Maximum number of aliases to resolve before giving up.
	#define MAXCHAINS 10
	
	OSErr		theErr;
	FileSpec	initFSSpec;
	Handle		alisHandle;
	SInt16		chainCount;
	SInt16		newResFile;
	SInt16		oldResFile;
	Boolean		needToClose;
	Boolean		updateFlag, foundFlag;
	Boolean		wasAliasedTemp, specChanged;
	
	if (spec == nil || targetIsFolder == nil || wasAliased == nil)
		return paramErr;
	
				// Init local vars.
	theErr = noErr;
	initFSSpec = *spec;
	chainCount = MAXCHAINS;
	newResFile = -1;
	oldResFile = CurResFile();
	*targetIsFolder = foundFlag = specChanged = false;
	
				// Loop through chain of alias files.
	do
	{
		chainCount--;
		
				// Check if FSSpec is an alias file or a directory.
				// Note that targetIsFolder => NOT wasAliased.
		theErr = MyIsAliasFile(spec, wasAliased, targetIsFolder);
		if (theErr != noErr || !(*wasAliased))
			break;
		
				// Open file's resource fork.
		newResFile = FSpOpenResFile(spec, fsCurPerm);
		require_action(newResFile != -1, bail, theErr=ResError(););
		
		needToClose = (CurResFile() == newResFile);
		
		UseResFile(newResFile);
		
				// The first 'alis' resource in the file is the appropriate alias.
		alisHandle = Get1IndResource(rAliasType, 1);
		require_action(alisHandle, bail, theErr=ResError(););
		
				// Load the resource explicitly in case SetResLoad(FALSE).
		LoadResource(alisHandle);
		theErr = ResError();
		require(theErr==noErr, bail);
		
				// FollowFinderAlias returns nsvErr if volume not mounted.
		theErr = FollowFinderAlias(spec, (AliasHandle)alisHandle, mountRemoteVols, spec, &updateFlag);
		require(theErr==noErr, bail);
		
		if (updateFlag)
		{
				// The resource in the alias file needs updating.
			ChangedResource(alisHandle);
			WriteResource(alisHandle);
		}
		
		specChanged = true;
		
		theErr = MyIsAliasFile(spec, &wasAliasedTemp, targetIsFolder);
		require(theErr==noErr, bail);
		
		foundFlag = !(wasAliasedTemp && resolveAliasChains);			
		
				// Restore the previous resource chain.
		if (newResFile != -1)
		{
			if (needToClose)
				CloseResFile(newResFile);
			UseResFile(oldResFile);
		}
		newResFile = -1;
	}
	while ( (theErr == noErr) && (chainCount > 0) && !foundFlag);
	
				// Return file not found error for circular alias chains.
	if (chainCount == 0 && !foundFlag)
		theErr = fnfErr;
	
bail:
				// Restore the previous resource chain.
	if (newResFile != -1)
	{
		if (needToClose)
			CloseResFile(newResFile);
		UseResFile(oldResFile);
	}
	
				// If error occurred restore the original FSSpec.
	if (theErr && specChanged)
		*spec = initFSSpec;
	
	return theErr;
}

//--------------------------------------------------------------------- FileClose

OSErr FileClose ( FileRef refNum )
{
	if (refNum)
	{
		ParamBlockRec	pb;
		
		pb.ioParam.ioCompletion = nil;
		pb.ioParam.ioRefNum = refNum;
		(void) PBFlushFileSync(&pb);
		
		return FSClose(refNum);
	}
	else
		return noErr;
}

//--------------------------------------------------------------------- FileGetEOF
//	Abstract:
//		Return the overall length of an open file.
//		
//	Params:
//		FileRef			(in) 	File's i/o refnum
//		long *			(out)	ptr to long to receive eof value
//	Return:
//		OSErr			if a file system error is encountered.
//---------------------------------------------------------------------
OSErr	FileGetEOF	(	FileRef			file,
						unsigned long	*eof )
{
	OSErr	theErr = fnOpnErr;
	
	// Might as well check the file param
	if (file)
	{
		theErr = GetEOF(file,(long*) eof);
	}

	// Return error
	return (theErr);
}

//--------------------------------------------------------------------- FileSetEOF
//	Abstract:
//		Reset the eof marker of an open file.
//		
//	Params:
//		FileRef			(in) 	File's i/o refnum
//		long			(in)	new eof value
//	Return:
//		OSErr			if a file system error is encountered.
//---------------------------------------------------------------------
OSErr	FileSetEOF	(		FileRef			file,
							unsigned long	eof )
{
	OSErr	theErr = fnOpnErr;
	
	// Might as well check the file param
	if (file)
	{
		theErr = SetEOF(file,eof);
	}

	// Return error
	return (theErr);
}

#pragma mark <<<<<< Static Functions >>>>>>

//===================================================================== Private Functions
//--------------------------------------------------------------------- MyIsAliasFile
// Determines if a file system object is a Finder alias. See Tech Note 315.
//	
//	fileFSSpec				the location of the file system object
//	aliasFileFlag			on exit, indicates whether the object was a Finder alias file
//	folderFlag				on exit, indicates whether the object was a directory
//	
//	returns					noErr if successful, or an error code
//--------------------------------------------------------------------- 
static OSErr MyIsAliasFile (const FileSpec *fileFSSpec, Boolean *aliasFileFlag, Boolean *folderFlag)
{
	CInfoPBRec	myCInfoPBRec;
	OSErr		retCode;
	
	if (fileFSSpec == nil || aliasFileFlag == nil || folderFlag == nil)
		return paramErr;
	
	*aliasFileFlag = *folderFlag = false;
	
				// Get the item's catalog information.
	myCInfoPBRec.hFileInfo.ioCompletion = nil;
	myCInfoPBRec.hFileInfo.ioNamePtr = (StringPtr) &fileFSSpec->name;
	myCInfoPBRec.hFileInfo.ioVRefNum = fileFSSpec->vRefNum;
	myCInfoPBRec.hFileInfo.ioDirID = fileFSSpec->parID;
	myCInfoPBRec.hFileInfo.ioFVersNum = 0;
	myCInfoPBRec.hFileInfo.ioFDirIndex = 0;
	
	retCode = PBGetCatInfoSync(&myCInfoPBRec);
	
				// Set aliasFileFlag if the item is not a directory and the
				// aliasFile bit is set.
	if (retCode == noErr)
	{
				// Check directory bit.
		if ((myCInfoPBRec.hFileInfo.ioFlAttrib & ioDirMask) != 0)
		{
			*folderFlag = true;
		}		// Check isAlias bit.
		else if ((myCInfoPBRec.hFileInfo.ioFlFndrInfo.fdFlags & kIsAlias) != 0)
		{
			*aliasFileFlag = true;
		}
	}
	
	return retCode;
}

//--------------------------------------------------------------------- CopyFile
//	Abstract:
//		Copied from MoreFileExtras.c
//		Copy an entire fork of a file from the source to the destination 
//		using copyBuffer as the buffer.
//		
//	Params:
//		srcRefNum		(in)	Source file refrence number
//		dstRefNum		(in)	Destination file refrence number
//		copyBufferPtr	(in)	Buffer to use when copying
//		copyBufferSize	(in)	Size of copy buffer
//		
//	Return:
//		noErr					successful
//		FileMgr Error
//---------------------------------------------------------------------

static OSErr CopyFile (	FileRef		srcRefNum,
						FileRef		dstRefNum,
						void		*copyBufferPtr,
						SInt32		copyBufferSize )
{
	ParamBlockRec	srcPB;
	ParamBlockRec	dstPB;
	OSErr			srcError;
	OSErr			dstError;
	UInt32			size;
	
	if ((copyBufferPtr == nil) || (copyBufferSize == 0))
		return (paramErr);
	
	/* preallocate the destination fork and */
	/* ensure the destination fork's EOF is correct after the copy */
	srcError = FileGetEOF( srcRefNum, &size );
	if (srcError != noErr)
		return (srcError);
	dstError = FileSetEOF( srcRefNum, size );
	if (dstError != noErr)
		return (dstError);
	
	srcPB.ioParam.ioRefNum = srcRefNum;
	dstPB.ioParam.ioRefNum = dstRefNum;
	
				// Reset source fork's mark.
	srcPB.ioParam.ioPosMode = fsFromStart;
	srcPB.ioParam.ioPosOffset = 0;
	srcError = PBSetFPosSync(&srcPB);
	if (srcError != noErr)
		return (srcError);
	
				// Reset destination fork's mark.
	dstPB.ioParam.ioPosMode = fsFromStart;
	dstPB.ioParam.ioPosOffset = 0;
	dstError = PBSetFPosSync(&srcPB);
	if (dstError != noErr)
		return (dstError);
	
				// Set up fields that won't change in the loop.
	srcPB.ioParam.ioBuffer = (Ptr)copyBufferPtr;
	srcPB.ioParam.ioPosMode = fsAtMark + kNoCacheBit;
	srcPB.ioParam.ioReqCount = ((copyBufferSize >= 512) && (copyBufferSize % 512)) ? (copyBufferSize / 512) * 512 : (copyBufferSize);
	/* If copyBufferSize is greater than 512 bytes, make it a multiple of 512 bytes */
	/* This will make writes on local volumes faster */
	
	dstPB.ioParam.ioBuffer = (Ptr)copyBufferPtr;
	dstPB.ioParam.ioPosMode = fsAtMark + kNoCacheBit;
	
	while ((srcError == noErr) && (dstError == noErr))
	{
		srcError = PBReadSync(&srcPB);
		dstPB.ioParam.ioReqCount = srcPB.ioParam.ioActCount;
		dstError = PBWriteSync(&dstPB);
	}
	
				// Make sure there were no errors at the destination.
	if (dstError != noErr)
		return (dstError);
	
				// Make sure the only error at the source was eofErr.
	if (srcError != eofErr)
		return (srcError);
	
	return (noErr);
}

