/*
	File:		GIF.c

	Contains:	

	Version:	ColorSync 2 or later

	Copyright:	 1998 by Apple Computer, Inc., all rights reserved.

	File Ownership:

		DRI:				David Hayward

		Other Contact:		Steve Swen

		Technology:			ColorSync 2

	Writers:

		(DH)	David Hayward

	Change History (most recent first):

		 <1>	 3/26/98	DH		First checked in.
*/

#include <string.h>

#include <Types.h>
#include <Files.h>
#include <CMApplication.h>
#include <CMICCProfile.h>
#include <CMScriptingPlugin.h>

#define COMPONENT_SIGNATURE 'GIF '
#include "DebugAssert.h"

#include "plugUtil.h"
#include "plugIO.h"
#include "FileUtils.h"
#include "GIFcore.h"


CMError ValidateImage	(const FSSpec *spec);
CMError GetImageSpace	(const FSSpec *spec, OSType *space);
CMError ValidateSpace	(const FSSpec *spec, OSType *space);
CMError EmbedImage		(const FSSpec *inSpec, const FSSpec *outSpec, CMProfileRef embedProf, UInt32 embedFlags);
CMError MatchImage		(const FSSpec *inSpec, const FSSpec *outSpec, UInt32 qual, UInt32 srcIntent, 
						 CMProfileRef srcProf, CMProfileRef dstProf, CMProfileRef prfProf, UInt32 matchFlags);


CMError ValidateImage	(const FSSpec *spec)
{
	CMError				err = noErr;
	FILE*				in = 0;
	char				head[6];
	
	in = fopen_spec(spec, "r");
	require_action( in, bail, err=ioErr; );
	
	err = ReadSize(in, sizeof(head), (Ptr)head );
	require( err==noErr, bail);
	
 	err = cmspInvalidImageFile;
	if (memcmp(head,"GIF89a",sizeof(head))==0 ||
		memcmp(head,"GIF87a",sizeof(head))==0)
		err = noErr;
	
bail:
	
	if (in) fclose(in);
	
	return err;
}


CMError GetImageSpace	(const FSSpec *spec, OSType *space)
{
	*space = cmRGBData;
	return noErr;
}


CMError ValidateSpace	(const FSSpec *spec, OSType *space)
{
	if (*space == cmRGBData)
		return noErr;
	else
		return cmspInvalidImageSpace;
}


CMError EmbedImage		(const FSSpec *inSpec, const FSSpec *outSpec, CMProfileRef embedProf, UInt32 embedFlags)
{
	CMError				err = noErr;
	FILE*				in = 0;
	FILE*				out = 0;
	
	in = fopen_spec(inSpec, "r");
	require_action( in, bail, err=ioErr; );
	
	out = fopen_spec(outSpec, "w");
	require_action( out, bail, err=ioErr; );
	
	err = GIFEmbed(in, out, embedProf );
	require( err==noErr, bail);
	
	(void) RsrcCopy( inSpec, outSpec);
	(void) InfoCopy( inSpec, outSpec);
	
bail:
	
	if (in) fclose(in);
	if (out) fclose(out);
	
	return err;
}



CMError MatchImage		(const FSSpec *inSpec, const FSSpec *outSpec, UInt32 qual, UInt32 srcIntent, 
						 CMProfileRef srcProf, CMProfileRef dstProf, CMProfileRef prfProf, UInt32 matchFlags)
{
	CMError				err = noErr;
	FILE*				in = 0;
	FILE*				out = 0;
	CMProfileRef		firstProf = nil;
	CMWorldRef			cw = nil;	
	
	in = fopen_spec(inSpec, "r");
	require_action( in, bail, err=ioErr; );
	
	out = fopen_spec(outSpec, "w");
	require_action( out, bail, err=ioErr; );
	
				// determine 1st profile in colorworld
	if (matchFlags & cmspFavorEmbeddedMask)
		(void) GIFOpenEmbeddedProfile(in, &firstProf,  qual, srcIntent);
	if (!firstProf)
		firstProf = srcProf;
	
				// build the colorworld
	err = PrepareColorWorld( firstProf, dstProf, prfProf, &cw );
	if (firstProf != srcProf)
		CMCloseProfile(firstProf);
	require( err==noErr, bail);
	
	err = GIFMatch(in, out, prfProf?prfProf:dstProf, cw );
	require( err==noErr, bail);
	
	(void) RsrcCopy( inSpec, outSpec);
	(void) InfoCopy( inSpec, outSpec);
	
bail:
	
	if (in)   fclose(in);
	if (out)  fclose(out);
	if (cw)   CWDisposeColorWorld(cw);
	
	return err;
}
