/*
	File:		GIFcore.c
	Contains:	
	Version:	ColorSync 2 or later
	Copyright:	 1998 by Apple Computer, Inc., all rights reserved.
	
	File Ownership:
	
		DRI:				David Hayward
		Other Contact:		Steve Swen
		Technology:			ColorSync 2
	
	Writers:

		(ebb)	Eric Broadbent	
		(DH)	David Hayward
	
	Change History (most recent first):

		<17>	 04/29/98	WN		SmartSetHandleSize / SmartSetPtrSize parameter changes
		<16>	 4/28/98	ebb		Use SmartMem routines, defined in MemUtils.h.
		<15>	 04/29/98	WN		SmartSetHandleSize / SmartSetPtrSize parameter changes
		 <3>	 4/10/98	DH		Various improvements.
		 <2>	 3/26/98	DH		Dont embed profile if nil.
		 <1>	 3/26/98	DH		First checked in.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <ToolUtils.h>
#include <OSUtils.h>
#include <Palettes.h>
#include <Displays.h>
#include "DebugAssert.h"

#include "GIFcore.h"
#include "MemUtils.h"
#include "FileUtils.h"
#include "plugIO.h"
#include "plugUtil.h"


/* Local function prototype */
static UInt16 ColorsFromPackedField (UInt8	packedField);

static OSErr CopyGifHeader				(FILE *, FILE *, GifHeaderPtr );
static OSErr CopyGifImageDesc			(FILE *, FILE *, GifImageDescPtr );
static OSErr CopyGifColorTable			(FILE *, FILE *, Ptr, UInt16 );
static OSErr MatchGifColorTable			(FILE *, FILE *, UInt16, CMWorldRef );
static OSErr CopyGifGraphicControl		(FILE *, FILE *, GifGraphicControlPtr );
static OSErr CopyGifPlainText			(FILE *, FILE *, GifPlainTextPtr );
static OSErr CopyGifApplication			(FILE *, FILE *, GifApplicationPtr );
static OSErr CopyGifComment				(FILE *, FILE *, GifCommentPtr );
static OSErr CopyDataSubBlocks			(FILE *, FILE *, Handle );

static OSErr DoEmbedProfile				(FILE *, CMProfileRef embedProf);
static Boolean GifApplicationIsProfile	(GifApplicationPtr g );






OSErr GIFEmbed ( FILE *from, FILE *into, CMProfileRef embedProf)
{
	return GIFMatch(from, into, embedProf, nil);
}


OSErr GIFMatch ( FILE *from, FILE *into, CMProfileRef embedProf, CMWorldRef cw)
{
	OSErr					theErr = noErr;
	UInt8					Identifier;			/* Extension block identifier holder */
	UInt8					Label;				/* Extension block label holder */
	UInt16					tableSize;			/* Number of entries in the Local or Global Color Table */
	GifHeaderRec			gifHead;			/* GIF Header structure */
	GifImageDescRec			gifImageDesc;		/* Logical Image Descriptor struct */
	GifApplicationRec		gifApplication;		/* Application Extension structure */
	Ptr						colors = nil;
	
				// Rewind to beginning of file
	theErr = RewindFile(from);
	require (theErr==noErr, bail);
	
	theErr = RewindFile(into);
	require (theErr==noErr, bail);
	
				// Copy the GIF image file header information
	theErr = CopyGifHeader(from, nil, &gifHead);
	require (theErr==noErr, bail);
	
	if (gifHead.Version[1] == '7')
		gifHead.Version[1] = '9';
	
	theErr = CopyGifHeader(nil, into, &gifHead);
	require (theErr==noErr, bail);

				// Copy the Global Color List if present 
	tableSize = ColorsFromPackedField(gifHead.PackedField);
	if (tableSize)
	{
		theErr = MatchGifColorTable(from, into, tableSize, cw );
		require (theErr==noErr, bail);
	}
	
				// Embed the profile in the file
	theErr = DoEmbedProfile(into, embedProf);
	require (theErr==noErr, bail);
	
				// Loop through file
	while (theErr==noErr)
	{
		theErr = ReadByte(from, &Identifier);
		require (theErr==noErr, bail);
		
		switch (Identifier)
		{
			case kGifImageSeparator:
				gifImageDesc.ImageSeparator = Identifier;
				theErr = CopyGifImageDesc(from, into, &gifImageDesc);
				require (theErr==noErr, bail);
				
					// Copy the Local Color List if present 
				tableSize = ColorsFromPackedField(gifImageDesc.PackedField);
				if (tableSize)
				{
					theErr = MatchGifColorTable(from, into, tableSize, cw );
					require (theErr==noErr, bail);
				}
	
					// Copy LZW Minimum Code Size Value
				theErr = CopyByte(from, into, nil);
				require (theErr==noErr, bail);
				
					// Copy encoded image data
				theErr = CopyDataSubBlocks(from, into, nil );
				require (theErr==noErr, bail);
				break;
				
			case kGifExtensionIntroducer:
				theErr = ReadByte(from, &Label);
				require (theErr==noErr, bail);
				
				switch (Label)
				{
					case kGifPlainTextExtLabel:
						theErr = CopyGifPlainText(from, into, nil);
						check(theErr==noErr);
						break;
						
					case kGifCommentExtLabel:
						theErr = CopyGifComment(from, into, nil);
						check(theErr==noErr);
						break;
						
					case kGifGraphicControlExtLabel:
						theErr = CopyGifGraphicControl(from, into, nil);
						check(theErr==noErr);
						break;
						
					case kGifApplicationExtLabel:
						gifApplication.data = SmartNewHandleClear( 256, &theErr);
						require( theErr==noErr, bail);
	
						theErr = CopyGifApplication(from, nil, &gifApplication);
						require (theErr==noErr, bail);
						
						if (!GifApplicationIsProfile(&gifApplication))
						{
							theErr = CopyGifApplication(nil, into, &gifApplication);
							require (theErr==noErr, bail);
						}
						SmartDisposeHandle(&(gifApplication.data));
						break;
						
					default:
						theErr = paramErr; // Unknown Extension Label
						check(theErr==noErr);
						break;
				}
				break;
				
			case kGifTrailer:
				theErr = CopyByte(nil, into, &Identifier);
				require( theErr==noErr, bail);
				goto bail; // we're done
				break;
				
			default:
				theErr = paramErr; // Unknown Block Separator Character
				check(theErr==noErr);
				break;
		}
	}
	
bail:
	
	SmartDisposePtr(&colors);
	
	return theErr;
}

CMError GIFOpenEmbeddedProfile ( FILE* from, CMProfileRef *ref, UInt32 qual, UInt32 intent )
{
	OSErr					theErr = noErr;
	UInt8					Identifier;			/* Extension block identifier holder */
	UInt8					Label;				/* Extension block label holder */
	UInt16					tableSize;			/* Number of entries in the Local or Global Color Table */
	GifHeaderRec			gifHead;			/* GIF Header structure */
	GifImageDescRec			gifImageDesc;		/* Logical Image Descriptor struct */
	GifApplicationRec		gifApplication;		/* Application Extension structure */
	
	*ref = nil;
	
				// Rewind to beginning of file
	theErr = RewindFile(from);
	require (theErr==noErr, bail);
	
				// Read the GIF image file header information
	theErr = CopyGifHeader(from, nil, &gifHead);
	require (theErr==noErr, bail);
	
				// Read the Global Color List if present 
	tableSize = ColorsFromPackedField(gifHead.PackedField);
	if (tableSize)
	{
		theErr = CopyGifColorTable(from, nil, nil, tableSize);
		require (theErr==noErr, bail);
	}
	
				// Loop through file
	while (theErr==noErr && (*ref)==nil)
	{
		theErr = ReadByte(from, &Identifier);
		require (theErr==noErr, bail);
		
		switch (Identifier)
		{
			case kGifImageSeparator:
				gifImageDesc.ImageSeparator = Identifier;
				theErr = CopyGifImageDesc(from, nil, &gifImageDesc);
				require (theErr==noErr, bail);
				
					// Read the Local Color List if present 
				tableSize = ColorsFromPackedField(gifImageDesc.PackedField);
				if (tableSize)
				{
					theErr = CopyGifColorTable(from, nil, nil, tableSize);
					require (theErr==noErr, bail);
				}
	
					// Read LZW Minimum Code Size Value
				theErr = CopyByte(from, nil, nil);
				require (theErr==noErr, bail);
				
					// Read encoded image data
				theErr = CopyDataSubBlocks(from, nil, nil );
				require (theErr==noErr, bail);
				break;
				
			case kGifExtensionIntroducer:
				theErr = ReadByte(from, &Label);
				require (theErr==noErr, bail);
				
				switch (Label)
				{
					case kGifPlainTextExtLabel:
						theErr = CopyGifPlainText(from, nil, nil);
						check(theErr==noErr);
						break;
						
					case kGifCommentExtLabel:
						theErr = CopyGifComment(from, nil, nil);
						check(theErr==noErr);
						break;
						
					case kGifGraphicControlExtLabel:
						theErr = CopyGifGraphicControl(from, nil, nil);
						check(theErr==noErr);
						break;
						
					case kGifApplicationExtLabel:
						gifApplication.data = SmartNewHandleClear( 256, &theErr);
						require( theErr==noErr, bail);
	
						theErr = CopyGifApplication(from, nil, &gifApplication);
						require (theErr==noErr, bail);
						
						if (GifApplicationIsProfile(&gifApplication))
						{
							theErr = OpenProfileHdlAsTemp(gifApplication.data, ref, qual, intent );
							require( theErr==noErr, bail);
						}
						
						SmartDisposeHandle(&(gifApplication.data));
						break;
						
					default:
						theErr = paramErr; // Unknown Extension Label
						check(theErr==noErr);
						break;
				}
				break;
				
			case kGifTrailer:
				theErr = CopyByte(nil, nil, &Identifier);
				require( theErr==noErr, bail);
				goto bail;
				return noErr;
				
			default:
				theErr = paramErr; // Unknown Block Separator Character
				check(theErr==noErr);
				break;
		}
	}
	
bail:
	
	if (*ref == nil)
		theErr = cmProfileNotFound;
	
	return theErr;
}


#pragma mark -

static UInt16 ColorsFromPackedField (UInt8	packedField)
{
	if (packedField & 0x80)
		return (UInt16) (1L << ((packedField & 0x07) + 1));
	else
		return 0;
}

/*
**  Copy a GIF image file Header.
**
**  This function reads the Header, Logical Screen Descriptor, and
**  Global Color Table (if any) from a GIF image file.  The information
**  is stored in a GifHeaderRec structure.
**
**  Returns: -1 if a FILE stream error occured during the read,
**			otherwise 0 if no error occured.
*/
static OSErr CopyGifHeader (FILE *from, FILE *into, GifHeaderPtr g)
{
	OSErr		theErr = noErr;
	
	theErr = CopySize(from, into, sizeof(g->Signature), g ? ((Ptr)g->Signature) : nil );
	require( theErr==noErr, bail);
	
	theErr = CopySize(from, into, sizeof(g->Version), g ? ((Ptr)g->Version) : nil );
	require( theErr==noErr, bail);
	
	theErr = CopyWord(from, into, g ? (&(g->ScreenWidth)) : nil, kEndianL);
	require( theErr==noErr, bail);
	
	theErr = CopyWord(from, into, g ? (&(g->ScreenHeight)) : nil, kEndianL);
	require( theErr==noErr, bail);
	
	theErr = CopyByte(from, into, g ? (&(g->PackedField)) : nil);
	require( theErr==noErr, bail);
	
	theErr = CopyByte(from, into, g ? (&(g->ColorIndex)) : nil);
	require( theErr==noErr, bail);
	
	theErr = CopyByte(from, into, g ? (&(g->AspectRatio)) : nil);
	require( theErr==noErr, bail);
	
bail:
	
	return theErr;
}


/*
**  Copy a GIF Local Image Descriptor.
**
**  This function reads the Local Image Descriptor, and Local Color
**  Table (if any) from a GIF image file.  The information is stored
**  in a GifImageDescRec structure.
**
**  Note that the ImageSeparator field value in the GifImageDescRec
**  structure is assigned by the function calling CopyGifImageDesc().
**
**  Returns: -1 if a FILE stream error occured during the read,
**			otherwise 0 if no error occured.
*/
static OSErr CopyGifImageDesc (FILE *from, FILE *into, GifImageDescPtr g)
{
	OSErr		theErr = noErr;
	UInt8		ImageSeparator = kGifImageSeparator;
	
	theErr = CopyByte(nil, into, &ImageSeparator);
	require( theErr==noErr, bail);
	
	
	theErr = CopyWord(from, into, g ? (&(g->ImageLeft)) : nil, kEndianL);
	require( theErr==noErr, bail);
	
	theErr = CopyWord(from, into, g ? (&(g->ImageTop)) : nil, kEndianL);
	require( theErr==noErr, bail);
	
	theErr = CopyWord(from, into, g ? (&(g->ImageWidth)) : nil, kEndianL);
	require( theErr==noErr, bail);
	
	theErr = CopyWord(from, into, g ? (&(g->ImageHeight)) : nil, kEndianL);
	require( theErr==noErr, bail);
	
	theErr = CopyByte(from, into, g ? (&(g->PackedField)) : nil);
	require( theErr==noErr, bail);
	
bail:
	
	return theErr;
}


/*
**  Copy a GIF Color Talbe.
**
**  This function reads the Local or Global Color Table from a GIF image file.
**  The information is stored in a GifColorTable structure.
**
**  Returns: readErr, writErr, eofErr or noErr
*/
static OSErr CopyGifColorTable (FILE *from, FILE *into, Ptr c, UInt16 count)
{
	OSErr		theErr = noErr;
	UInt32		size;
	
	size = count * 3;

	if (from && c)				// grow data buffer
	{
		SetPtrSize(c, size);
		theErr = MemError();
		require( theErr==noErr, bail);
	}
	
	theErr = CopySize(from, into, size, (Ptr)c );
	require( theErr==noErr, bail);
	
bail:
	
	return theErr;
}


static OSErr MatchGifColorTable (FILE *from, FILE *into, UInt16 count, CMWorldRef cw )
{
	OSErr		theErr = noErr;
	Ptr			colors = nil;
	CMBitmap	bitmap;
	
	require_action( from && into, bail, theErr=paramErr; );
	
	if (count && cw==nil)
	{
		theErr = CopyGifColorTable(from, into, nil, count);
		require (theErr==noErr, bail);
	}
	else if (count)
	{
		colors = SmartNewPtr(count*3, &theErr);
		require (theErr==noErr, bail);
		
		theErr = CopyGifColorTable(from, nil, colors, count);
		require (theErr==noErr, bail);
		
				// set up bitmap to match gif colors
		bitmap.image		= (char*)colors;
		bitmap.width		= count;
		bitmap.height		= 1;
		bitmap.rowBytes		= count*3;
		bitmap.pixelSize	= 24;
		bitmap.space		= cmRGB24Space;
		bitmap.user1		= 0;
		bitmap.user2		= 0;
		
				// match colors in place
		theErr = CWMatchBitmap(cw, &bitmap, nil, nil, nil);
		require (theErr==noErr, bail);
		
		theErr = CopyGifColorTable(nil, into, colors, count);
		require (theErr==noErr, bail);
	}
	
bail:
	
	SmartDisposePtr(&colors);
	
	return theErr;
}


/*
**  Copy a GIF Graphic Control Extension block.
**
**  Note that the Introducer and Label field values in the GifGraphicControlRec
**  structure are assigned by the function calling CopyGifGraphicControl().
**
**  Returns: readErr, writErr, eofErr or noErr
*/
static OSErr CopyGifGraphicControl (FILE *from, FILE *into, GifGraphicControlPtr g )
{
	OSErr		theErr = noErr;
	UInt8		Introducer = kGifExtensionIntroducer;
	UInt8		Label = kGifGraphicControlExtLabel;
	
	theErr = CopyByte(nil, into, &Introducer);
	require( theErr==noErr, bail);
	
	theErr = CopyByte(nil, into, &Label);
	require( theErr==noErr, bail);
	
	
	theErr = CopyByte(from, into, g ? (&(g->BlockSize)) : nil);
	require( theErr==noErr, bail);
	
	theErr = CopyByte(from, into, g ? (&(g->PackedField)) : nil);
	require( theErr==noErr, bail);
	
	theErr = CopyWord(from, into, g ? (&(g->DelayTime)) : nil, kEndianL);
	require( theErr==noErr, bail);
	
	theErr = CopyByte(from, into, g ? (&(g->ColorIndex)) : nil);
	require( theErr==noErr, bail);
	
				// Copy in the Graphic Control data sub-blocks
	theErr = CopyDataSubBlocks(from, into, g ? (g->data) : nil );
	require( theErr==noErr, bail);
	
bail:
	
	return theErr;
}


/*
**  Copy a GIF Plain Text Extension block.
**
**  Note that the Introducer and Label field values in the GifPlainTextRec
**  structure are assigned by the function calling CopyGifPlainText().
**
**  Returns: readErr, writErr, eofErr or noErr
*/
static OSErr CopyGifPlainText (FILE *from, FILE *into, GifPlainTextPtr g )
{
	OSErr		theErr = noErr;
	UInt8		Introducer = kGifExtensionIntroducer;
	UInt8		Label = kGifPlainTextExtLabel;
	
	theErr = CopyByte(nil, into, &Introducer);
	require( theErr==noErr, bail);
	
	theErr = CopyByte(nil, into, &Label);
	require( theErr==noErr, bail);
	
	
	theErr = CopyByte(from, into, g ? (&(g->BlockSize)) : nil);
	require( theErr==noErr, bail);
	
	theErr = CopyWord(from, into, g ? (&(g->TextGridLeft)) : nil, kEndianL);
	require( theErr==noErr, bail);
	
	theErr = CopyWord(from, into, g ? (&(g->TextGridTop)) : nil, kEndianL);
	require( theErr==noErr, bail);
	
	theErr = CopyWord(from, into, g ? (&(g->TextGridWidth)) : nil, kEndianL);
	require( theErr==noErr, bail);
	
	theErr = CopyWord(from, into, g ? (&(g->TextGridHeight)) : nil, kEndianL);
	require( theErr==noErr, bail);
	
	theErr = CopyByte(from, into, g ? (&(g->CellWidth)) : nil);
	require( theErr==noErr, bail);
	
	theErr = CopyByte(from, into, g ? (&(g->CellHeight)) : nil);
	require( theErr==noErr, bail);
	
	theErr = CopyByte(from, into, g ? (&(g->TextFgColorIndex)) : nil);
	require( theErr==noErr, bail);
	
	theErr = CopyByte(from, into, g ? (&(g->TextBgColorIndex)) : nil);
	require( theErr==noErr, bail);
	
				// Copy in the Plain Text data sub-blocks
	theErr = CopyDataSubBlocks(from, into, g ? (g->data) : nil);
	require( theErr==noErr, bail);
	
bail:
	
	return theErr;
}


/*
**  Copy a GIF Application Extension block.
**
**  Note that the Introducer and Label field values in the GifApplicationRec
**  structure are assigned by the function calling CopyGifApplication().
**
**  Returns: readErr, writErr, eofErr or noErr
*/
static OSErr CopyGifApplication (FILE *from, FILE *into, GifApplicationPtr g )
{
	OSErr		theErr = noErr;
	UInt8		Introducer = kGifExtensionIntroducer;
	UInt8		Label = kGifApplicationExtLabel;
	
	theErr = CopyByte(nil, into, &Introducer);
	require( theErr==noErr, bail);
	
	theErr = CopyByte(nil, into, &Label);
	require( theErr==noErr, bail);
	
	
	theErr = CopyByte(from, into, g ? (&(g->BlockSize)) : nil);
	require( theErr==noErr, bail);
	
	theErr = CopySize(from, into, sizeof(g->Identifier), g ? ((Ptr)g->Identifier) : nil);
	require( theErr==noErr, bail);
	
	theErr = CopySize(from, into, sizeof(g->AuthentCode), g ? ((Ptr)g->AuthentCode) : nil);
	require( theErr==noErr, bail);
	
				// Copy in the Plain Text data sub-blocks
	theErr = CopyDataSubBlocks(from, into, g ? (g->data) : nil);
	require( theErr==noErr, bail);
	
bail:
	
	return theErr;
}


/*
**  Copy a GIF Comment Extension block.
**
**  Note that the Introducer and Label field values in the GifCommentRec
**  structure are assigned by the function calling CopyGifComment().
**
**  Returns: readErr, writErr, eofErr or noErr
*/
static OSErr CopyGifComment (FILE *from, FILE *into, GifCommentPtr g)
{
	OSErr		theErr = noErr;
	UInt8		Introducer = kGifExtensionIntroducer;
	UInt8		Label = kGifCommentExtLabel;
	
	theErr = CopyByte(nil, into, &Introducer);
	require( theErr==noErr, bail);
	
	theErr = CopyByte(nil, into, &Label);
	require( theErr==noErr, bail);
	
	
				// Copy in the Plain Text data sub-blocks
	theErr = CopyDataSubBlocks(from, into, g ? (g->data) : nil);
	require( theErr==noErr, bail);
	
bail:
	
	return theErr;
}


/*
**  Copy one or more GIF data sub-blocks and write the information
**  to a buffer.
**
**  A GIF "sub-block" is a single count byte followed by 1 to 255
**  additional data bytes.
**
**  Returns: readErr, writErr, eofErr or noErr
*/
static OSErr CopyDataSubBlocks (FILE *from, FILE *into, Handle h )
{
	OSErr		theErr = noErr;
	UInt32		bufSize = 0;		/* Total size of the Plain Text data buffer */
	UInt8		subSize;			/* Size of the current data sub-block being read */
	UInt32		hdlSize;
	char		tmpBuf[255];
	
	require_action(from || h, bail, theErr=paramErr; );
	
	if (!from && h)
	{
		hdlSize = SmartGetHandleSize(h);
		subSize = (hdlSize>255)?(255):(hdlSize) ;
		hdlSize -= subSize;
	}
	
				// do the size of the first sub-block
	theErr = CopyByte(from, into, &subSize);
	
	while (subSize && theErr==noErr)
	{
		if (from && h)				// grow data buffer
		{
			theErr = SmartSetHandleSize(&h, bufSize + subSize);
			if (theErr) break;
		}
		
		SmartLock(h);
		theErr = CopySize(from, into, subSize, h ? (Ptr)(*h + bufSize) : tmpBuf);
		SmartUnlock(h);
		if (theErr) break;
		
		bufSize += subSize;			// Running total of the buffer size
		
		if (!from && h)
		{
			subSize = (hdlSize>255)?(255):(hdlSize) ;
			hdlSize -= subSize;
		}
		
				// do the size of the next sub-block
		theErr = CopyByte(from, into, &subSize);
		if (theErr) break;
	}
	
bail:
	
	return theErr;
}


#pragma mark -


static OSErr DoEmbedProfile ( FILE *into, CMProfileRef embedProf)
{
	Handle					h = nil;
	CMProfileLocation		tempLoc;
	CMProfileRef			tempProf;
	OSErr					theErr = noErr;
	CMAppleProfileHeader	header;
	GifApplicationRec		gifApplication;		/* Application Extension structure */
	
	require_quiet(into && embedProf, bail);
	
	theErr = CMGetProfileHeader(embedProf, &header);
	require( theErr==noErr, bail);
	
	h = SmartNewHandleClear( header.cm2.size, &theErr);
	require( theErr==noErr, bail);
	
				// flatten profile into handle
	tempLoc.locType = cmHandleBasedProfile;
	tempLoc.u.handleLoc.h = h;
	theErr = CMCopyProfile(&tempProf, &tempLoc, embedProf);
	require( theErr==noErr, bail);
	
	CMCloseProfile(tempProf);
	
	gifApplication.Introducer		= kGifExtensionIntroducer;
	gifApplication.Label			= kGifApplicationExtLabel;
	gifApplication.BlockSize		= 0x0B;
	gifApplication.Identifier[0]	= 'I';
	gifApplication.Identifier[1]	= 'C';
	gifApplication.Identifier[2]	= 'C';
	gifApplication.Identifier[3]	= 'R';
	gifApplication.Identifier[4]	= 'G';
	gifApplication.Identifier[5]	= 'B';
	gifApplication.Identifier[6]	= 'G';
	gifApplication.Identifier[7]	= '1';
	gifApplication.AuthentCode[0]	= '0';
	gifApplication.AuthentCode[1]	= '1';
	gifApplication.AuthentCode[2]	= '2';
	gifApplication.data				= h;

	theErr = CopyGifApplication(nil, into, &gifApplication);
	require (theErr==noErr, bail);
	
bail:
	
	SmartDisposeHandle(&h);
	
	return theErr;
}

static Boolean GifApplicationIsProfile ( GifApplicationPtr g )
{
	return  (strncmp( (char*)(g->Identifier), "ICCRGBG1012", 11) == 0);
}

