/*
	File:		MemUtils.c

	Contains:	Memory Utilities for ColorSync

	Copyright:	 1997 by Apple Computer, Inc., all rights reserved.

	Version:	ColorSync 2.5 or later

	File Ownership:

		DRI:				David Hayward

		Other Contact:		Steve Swen

		Technology:			ColorSync 2

	Writers:

		(ebb)	Eric Broadbent
		(jkc)	john  calhoun
		(DH)	David Hayward

	Change History (most recent first):

		<3>	 	04/29/98	WN		SmartSetHandleSize / SmartSetPtrSize parameter changes
		<2>	 	 4/27/98	ebb		Moved ChooseHeap from MacUtils.c.
		<1>		08/25/97	ebb		Created. Moved some routines from MacUtils.c
									Removed DisposeIf... there are Smart macros now.
									Use std library memset() instead of SetMem().
									

	To do:
		Win32 support.
		Implement Ptr such as to be abel to support GetPtrSize on unix.
		Add debugging utils.
		
   ___________________________________________________________________________________________ */

#include <MacTypes.h>
#include <Errors.h>
#include <MacMemory.h>
#include <Traps.h>

#include "MemUtils.h"


/* Reserve size for SmartNewHandle() and SmartNewPtr() */
#define kMemReserve	32767

/* Static Prototypes */
long	ChooseHeap	(	SIZE	bytesNeeded,
						SIZE	reserve );

/* Return values for ChooseHeap() */
enum
{
	noHeap = 0,
	currentHeap,
	systemHeap,
	tempHeap
};


/* ______________________________________________________________________

	Ptr 
	SmartNewPtrMax( SIZE maximumNeeded,
				OSErr *resultErr,
				SIZE	*amountAllocated )
					
	Abstract:
		Get the largest ptr we can from anywhere using SmartNewPtr.  Max
		return the amount (not greater than maximumNeeded in 
		amountAllocated).
				
	Params:
		maximumNeeded	(in)		Maximum to try and get
		resultErr		(out)		Error condtion memFullErr/noErr
		amountAllocated	(in)		Actual Amount Allocated
		
	Return:
		Pointer to block if successful
		Otherwise nil
		We guarantee that if nil, then an err is returned in resultErr and visa-versa.

 _____________________________________________________________________ */
#define kMinimumBufferSize			16

Ptr
SmartNewPtrMax	( SIZE 	maximumNeeded, 
				  OSErr *resultErr, 
				  SIZE  *amountAllocated )
{
	Ptr		p = nil;	
	OSErr	theErr = noErr;
	
	*amountAllocated = 0;
	
	while ( !p && !theErr )
	{
		p = SmartNewPtr(maximumNeeded, &theErr);
		
		if (theErr)
		{
			maximumNeeded /= 2;						// make a smaller size
			if (maximumNeeded > kMinimumBufferSize)	// if big enough then
				theErr = noErr;						// clear theErr to try again
		}
		else
			*amountAllocated = maximumNeeded;
	}
	
	if (!theErr && !p)	theErr = memFullErr;
	if (theErr && p )	p = nil;
	if (resultErr)		*resultErr = theErr;
	return p;
}

/* _________________________________________________________________________

	void
	SetMem(void *bytePtr, SIZE numBytes, unsigned char byteValue)

	Abstract:
		Fill block with specified byte value.

	Params:
		bytePtr					(in)		Address of block
		numBytes				(in)		SIZE of block
		byteValue				(in)		Fill value
		
	Return:
		None
		
   _________________________________________________________________________ */

void
SetMem(void *bytePtr, SIZE numBytes, unsigned char byteValue)
{
	register long i;
	register long count = numBytes;
	register unsigned char *ptr = (unsigned char *)bytePtr;
	register unsigned char value = byteValue;

	for (i = 0; i < count; ++i)
		*ptr++ = value;
}

/* _________________________________________________________________________

	Handle
	SmartNewHandle(SIZE numBytes, OSErr *resultCode)

	Abstract:
		Attempt to allocate relocatable block of specified size.
		
		Try to maintain kMemReserve bytes after allocation.
		
		Respect allocation guidelines for ColorSync 2.0
			allocate from current zone
			else if (current zone != system zone)
				allocate from system zone
			else
				allocate from Process Manager temp zone
				

	Params:
		numBytes				(in)		Number of bytes to allocate
		resultErr				(out)		if not nil, pointer to return
											Memory Manager result.
		
	Return:
		Handle to block if successful
		Otherwise nil
		We guarantee that if nil, then an err is returned in resultErr and visa-versa.
		
   _________________________________________________________________________ */

Handle
SmartNewHandle	( SIZE 	byteCount, 
				  OSErr *resultErr)
{
	Handle	h;
	short	whichHeap;
	OSErr	theErr;
	
	whichHeap = ChooseHeap(byteCount, kMemReserve);
	
	switch (whichHeap)
	{
		case noHeap:
			h = nil;
			theErr = memFullErr;
			break;
		
		case currentHeap:
			h = NewHandle(byteCount);
			theErr = MemError();
			break;
		
		case systemHeap:
			h = NewHandleSys(byteCount);
			theErr = MemError();
			break;
		
		case tempHeap:
			h = TempNewHandle(byteCount, &theErr);
			break;
		
		default:							/* Error in ChooseHeap() */
			h = nil;
			theErr = paramErr;
			break;
	}
	
	if (!theErr && !h)	theErr = memFullErr;
	if (theErr && h )	h = nil;
	if (resultErr)		*resultErr = theErr;
	return h;
}

/* _________________________________________________________________________

	Handle
	SmartNewHandleClear(SIZE byteCount, OSErr *resultErr)

	Abstract:
		Attempt to allocate relocatable block of specified size.
		
		Try to maintain kMemReserve bytes after allocation.
		
		Respect allocation guidelines for ColorSync 2.0
			allocate from current zone
			else if (current zone != system zone)
				allocate from system zone
			else
				allocate from Process Manager temp zone

		Fill block with 0x00 if allocation is successful.

	Params:
		numBytes				(in)		Number of bytes to allocate
		resultErr				(out)		if not nil, pointer to return
											Memory Manager result.
		
	Return:
		Handle to block if successful
		Otherwise nil
		We guarantee that if nil, then an err is returned in resultErr and visa-versa.
		
   _________________________________________________________________________ */

Handle
SmartNewHandleClear	( SIZE 	byteCount, 
					  OSErr *resultErr )
{
	Handle	h;
	short	whichHeap;
	OSErr	theErr;
	
	whichHeap = ChooseHeap(byteCount, kMemReserve);
	
	switch (whichHeap)
	{
		case noHeap:
			h = nil;
			theErr = memFullErr;
			break;
		
		case currentHeap:
			h = NewHandleClear(byteCount);
			theErr = MemError();
			break;
		
		case systemHeap:
			h = NewHandleSysClear(byteCount);
			theErr = MemError();
			break;
		
		case tempHeap:
			h = TempNewHandle(byteCount, &theErr);
			if ( h && !theErr )
			{
				SetMem( *h, byteCount, 0x00);
			//	memset(*h, (char) 0x00, byteCount);
			}
			break;
		
		default:							/* Error in ChooseHeap() */
			h = nil;
			theErr = paramErr;
			break;
	}
	
	if (!theErr && !h)	theErr = memFullErr;
	if (theErr && h )	h = nil;
	if (resultErr)		*resultErr = theErr;
	return h;
}

/* _________________________________________________________________________

	Ptr
	SmartNewPtr(SIZE byteCount, OSErr *resultErr);

	Abstract:
		Attempt to allocate nonrelocatable block of specified size.
		
		Try to maintain kMemReserve bytes after allocation.
		
		Respect allocation guidelines for ColorSync 2.0
			allocate from current zone
			else if (current zone != system zone)
				allocate from system zone

	Params:
		numBytes				(in)		Number of bytes to allocate
		resultErr				(out)		Memory Manager result
		
	Return:
		Pointer to block if successful
		Otherwise nil
		We guarantee that if nil, then an err is returned in resultErr and visa-versa.
		
   _________________________________________________________________________ */

Ptr
SmartNewPtr	( SIZE 	byteCount, 
			  OSErr *resultErr )
{
	Ptr		p;
	short	whichHeap;
	OSErr	theErr;
	
	whichHeap = ChooseHeap(byteCount, kMemReserve);
	
	switch (whichHeap)
	{
		case noHeap:
			p = nil;
			theErr = memFullErr;
			break;
		
		case currentHeap:
			p = NewPtr(byteCount);
			theErr = MemError();
			break;
		
		case systemHeap:
		case tempHeap:						/* System heap is growable into temp zone */
			p = NewPtrSys(byteCount);
			theErr = MemError();
			break;
		
		default:							/* Error in ChooseHeap() */
			p = nil;
			theErr = paramErr;
			break;
	}
	
	if (!theErr && !p)	theErr = memFullErr;
	if (theErr && p )	p = nil;
	if (resultErr)		*resultErr = theErr;
	return p;
}

/* _________________________________________________________________________

	Ptr
	SmartNewPtrClear(SIZE byteCount, OSErr *resultErr)

	Abstract:
		Attempt to allocate nonrelocatable block of specified size.
		
		Try to maintain kMemReserve bytes after allocation.
		
		Respect allocation guidelines for ColorSync 2.0
			allocate from current zone
			else if (current zone != system zone)
				allocate from system zone

		Fill block with 0x00 if allocation is successful.
		
	Params:
		numBytes				(in)		Number of bytes to allocate
		resultErr				(out)		Memory Manager result
		
	Return:
		Pointer to block if successful
		Otherwise nil
		We guarantee that if nil, then an err is returned in resultErr and visa-versa.
		
   _________________________________________________________________________ */

Ptr
SmartNewPtrClear	( SIZE 	byteCount, 
					  OSErr *resultErr )
{
	Ptr		p;
	short	whichHeap;
	OSErr	theErr;
	
	whichHeap = ChooseHeap(byteCount, kMemReserve);
	
	switch (whichHeap)
	{
		case noHeap:
			p = nil;
			theErr = memFullErr;
			break;
		
		case currentHeap:
			p = NewPtrClear(byteCount);
			theErr = MemError();
			break;
		
		case systemHeap:
		case tempHeap:						/* System heap is growable into temp zone */
			p = NewPtrSysClear(byteCount);
			theErr = MemError();
			break;
		
		default:							/* Error in ChooseHeap() */
			p = nil;
			theErr = paramErr;
			break;
	}
	
	if (!theErr && !p)	theErr = memFullErr;
	if (theErr && p )	p = nil;
	if (resultErr)		*resultErr = theErr;
	return p;
}

/* _________________________________________________________________________

	SmartGetPtrSize

	Abstract:
		Return the size of a non-relocatable block.
				
	Params:
		Ptr			p			(in)		The Ptr to return the size of
		
	Return:
		SIZE		the size of the block
		
   _________________________________________________________________________ */
SIZE	
SmartGetPtrSize		( Ptr	p )
{
	// If block is nil, return 0
	if (p == nil)
		return (0);
	
	// Simply call the Mac Memory Mgr
	return (GetPtrSize(p));
}

/* _________________________________________________________________________

	SmartSetPtrSize

	Abstract:
		Attempt to resize a non-relocatable block to the specified size.
				
	Params:
		Ptr			p			(in)		The Ptr to resize
		SIZE		newSize		(in)		The new size to set the Ptr to
		
	Return:
		OSErr		will be non-zero if the operation failed.
		
   _________________________________________________________________________ */
OSErr	
SmartSetPtrSize		( Ptr 	*p, 
					  SIZE	newSize )
{
	OSErr	err = noErr;
	
	if (p == nil)
		err = memPCErr;		/*Pointer Check failed*/
	else
	{
		SetPtrSize(*p,newSize);
		err = MemError();
	}
		
	// Alleviate the need to do additional error checking
	return (err);
}

/* _________________________________________________________________________

	SmartGetHandleSize

	Abstract:
		Return the size of a relocatable block.
				
	Params:
		Handle		h			(in)		The Handle to return the size of
		
	Return:
		SIZE		the size of the block
		
   _________________________________________________________________________ */
SIZE	
SmartGetHandleSize		( Handle h )
{
	// If block is nil, return 0
	if (h == nil)
		return (0);
	
	// Simply call the Mac Memory Mgr
	return (GetHandleSize(h));
}

/* _________________________________________________________________________

	SmartSetHandleSize

	Abstract:
		Attempt to resize a relocatable block to the specified size.
				
	Params:
		Handle		*h			(in)		The Handle to resize
		SIZE		newSize		(in)		The new size to set the Handle to
		
	Return:
		OSErr		will be non-zero if the operation failed.
		
   _________________________________________________________________________ */
OSErr	
SmartSetHandleSize	( Handle *h, 
					  SIZE 	 newSize )
{
	OSErr	err = noErr;
	
	if (*h == nil)
		err = nilHandleErr;
	else
	{
		SetHandleSize(*h,newSize);
		err = MemError();
	}
		
	// Alleviate the need to do additional error checking
	return (err);
}

/* _________________________________________________________________________

	SmartHandToHand

	Abstract:
		Attempt to duplicate a relocatable block.
				
	Params:
		Handle		*h			(in)		The handle to create
		Handle		original	(in)		The handle to copy
		
	Return:
		OSErr		error encountered
		
   _________________________________________________________________________ */
OSErr	
SmartHandToHand			( Handle	*h, 
						  Handle 	original)
{
	OSErr	err = noErr;
	Handle	copy = original;	// get a copy
	
	// Pre-set h to nil
	if (h != nil) *h = nil;

	// Make sure we have decent params
	if (original == nil || h == nil)
		err = paramErr;	

	if (!err)
	{
		err = HandToHand(&copy);
		
		// Set return param to the newly allocated copy
		if (!err) *h = copy;
	}

	// return the error
	return (err);
}

/* _________________________________________________________________________

	SmartDisposePtr

	Abstract:
		Attempt to delete a non-relocatable block.
				
	Params:
		Ptr			*p			(in)		Address of the Ptr to delete
		
	Return:
		The address is cleared
		
   _________________________________________________________________________ */
void	
SmartDisposePtr		( Ptr*	p )
{
	if (p != nil)
	{
		if (*p != nil)
		{
			DisposePtr((Ptr) *p);
			*p = nil;
		}
	}
}

/* _________________________________________________________________________

	SmartDisposeHdl

	Abstract:
		Attempt to delete a relocatable block.
				
	Params:
		Handle			*h		(in)		Address of the Handle to delete
		
	Return:
		The address is cleared
		
   _________________________________________________________________________ */
void	
SmartDisposeHdl		( Handle	*h )
{
	if (h != nil)
	{
		if (*h != nil)
		{
			DisposeHandle(*h);
			*h = nil;
		}
	}
}

/* _________________________________________________________________________

	SmartLock

	Abstract:
		Attempt to lock a relocatable block.
				
	Params:
		Handle			h		(in)		The Handle to lock
		
	Return:
		nothing
		
   _________________________________________________________________________ */
void	
SmartLock		( Handle	h )
{
	if (h != nil)
	{
		HLock(h);
	}
}

/* _________________________________________________________________________

	SmartUnlock

	Abstract:
		Attempt to lock a relocatable block.
				
	Params:
		Handle			h		(in)		The Handle to lock
		
	Return:
		nothing
		
   _________________________________________________________________________ */
void	
SmartUnlock		( Handle	h )
{
	if (h != nil)
	{
		HUnlock(h);
	}
}

/* _________________________________________________________________________

	SmartHGetState

	Abstract:
		Return the state (locked or unlocked) of a relocatable block.
				
	Params:
		Handle			h		(in)		The Handle to get the state of
		
	Return:
		SInt8			the state (locked or unlocked) of the given Handle
		
   _________________________________________________________________________ */
SInt8	
SmartHGetState		( Handle	h )
{
	SInt8	state = 0;
	
	if (h != nil)
	{
		state = HGetState(h);
	}
	
	return (state);
}

/* _________________________________________________________________________

	SmartHSetState

	Abstract:
		Set the state (locked or unlocked) of a relocatable block.
				
	Params:
		Handle			h		(in)		The Handle to set the state of
		
	Return:
		nothing

   _________________________________________________________________________ */
void	
SmartHSetState		( Handle	h,
					  SInt8		state )
{
	if (h != nil)
	{
		HSetState(h,state);
	}
}

/* _________________________________________________________________________

	long
	ChooseHeap(SIZE bytesNeeded, SIZE reserve);

	Abstract:
		Return heap with maximum free space, allowing specified reserve.
		Order of preference is
			current zone
			system zone
			temp zone
		
		If reserve cannot be maintained then return heap with greatest
		free space.

	Params:
		bytesNeeded				(in)		Number of bytes to allocate
		reserve					(in)		Number of bytes to leave in reserve
		
	Return:
		noHeap
		currentHeap
		systemHeap
		tempHeap
		
   _________________________________________________________________________ */
   
long	
ChooseHeap		( SIZE		bytesNeeded,
				  SIZE		reserve )
{
	short	whichHeap;
	SIZE	currZoneTotalFree;
	SIZE	sysZoneTotalFree;
	SIZE	tempZoneTotalFree;
	
	bytesNeeded += reserve;
	currZoneTotalFree = sysZoneTotalFree = tempZoneTotalFree = 0;
	
				// not as quick check (no compacting) to see if enough in curr heap
	currZoneTotalFree = MaxBlock();
	if ( (SmartMemError(kMemErrorParam) == noErr) && (currZoneTotalFree>=bytesNeeded) )
		return currentHeap;
	
				// not as quick check (no compacting) to see if enough in sys heap
	sysZoneTotalFree = MaxBlockSys();
	if ( (SmartMemError(kMemErrorParam)==noErr) && (sysZoneTotalFree>=bytesNeeded) )
		return systemHeap;
	
				// slow check to see if enough in temp heap
				// _OSDispatch is not implemented during early phase of startup
	//if (TrapAvailable(_OSDispatch))
	{
		SIZE grow;
		tempZoneTotalFree = TempMaxMem(&grow);
		if ( (SmartMemError(kMemErrorParam)==noErr) && (tempZoneTotalFree>=bytesNeeded) )
			return tempHeap;
	}
	
				// If we can't leave reserve, select the biggest heap
	if (currZoneTotalFree >= sysZoneTotalFree)
		whichHeap = (currZoneTotalFree >= tempZoneTotalFree) ? currentHeap : tempHeap;
	else
		whichHeap = (sysZoneTotalFree >= tempZoneTotalFree)  ? systemHeap  : tempHeap;
	
	return whichHeap;
}

