/*
	File:		GIFcore.c

	Contains:	

	Version:	ColorSync 2 or later

	Copyright:	 1998 by Apple Computer, Inc., all rights reserved.

	File Ownership:

		DRI:				David Hayward

		Other Contact:		Steve Swen

		Technology:			ColorSync 2

	Writers:

		(ebb)	Eric Broadbent
		(DH)	David Hayward

	Change History (most recent first):

		 <4>	 4/28/98	ebb		Include Files.h.
		 <3>	 4/28/98	ebb		XP sanitized.
		 <2>	 4/10/98	DH		Various improvements.
		 <1>	 3/26/98	DH		First checked in.
*/


#include <stdio.h>

#include <Errors.h>
#include <Endian.h>
#include <Files.h>
#include "DebugAssert.h"

#include "MemUtils.h"
#include "plugIO.h"

//----------------------------------------------------------------------
//
//	SetDefault
//
//----------------------------------------------------------------------
static OSErr SetDefault (short newVRefNum, long newDirID, short *oldVRefNum, long *oldDirID);
static OSErr SetDefault (short newVRefNum, long newDirID, short *oldVRefNum, long *oldDirID)
{
	OSErr	error;
	
	/* Get the current default volume/directory. */
	error = HGetVol(NULL, oldVRefNum, oldDirID);
	if ( error == noErr )
	{
		/* Set the new default volume/directory */
		error = HSetVol(NULL, newVRefNum, newDirID);
	}
	
	return ( error );
}

//----------------------------------------------------------------------
//
//	RestoreDefault
//
//----------------------------------------------------------------------
static OSErr RestoreDefault (short oldVRefNum, long oldDirID);
static OSErr RestoreDefault (short oldVRefNum, long oldDirID)
{
	OSErr	error;
	short	defaultVRefNum;
	long	defaultDirID;
	long	defaultProcID;
	
	/* Determine if the default volume was a wdRefNum. */
	error = GetWDInfo(oldVRefNum, &defaultVRefNum, &defaultDirID, &defaultProcID);
	if ( error == noErr )
	{
		/* Restore the old default volume/directory, one way or the other. */
		if ( defaultDirID != fsRtDirID )
		{
			/* oldVRefNum was a wdRefNum - use SetVol */
			error = SetVol(NULL, oldVRefNum);
		}
		else
		{
			/* oldVRefNum was a real vRefNum - use HSetVol */
			error = HSetVol(NULL, oldVRefNum, oldDirID);
		}
	}
	
	return ( error );
}

static char* Pstrcpy2C (char *dst, ConstStr255Param src, const unsigned char dstLen );
static char* Pstrcpy2C (char *dst, ConstStr255Param src, const unsigned char dstLen )
{
	unsigned char bytesToCopy ;
	
	bytesToCopy = *src ;
	if (bytesToCopy > dstLen)					// make sure were not too big
		bytesToCopy = dstLen ;
		
	BlockMoveData(	src + 1, dst,
					bytesToCopy ) ;				// copy string in
	dst[bytesToCopy] = 0 ;						// terminate c-string
	
	return dst ;
}
//----------------------------------------------------------------------
//
//	fopen_spec
//
//----------------------------------------------------------------------
FILE *fopen_spec(const FSSpec *spec, const char *mode)
{
	OSErr	error;
	FILE 	*fp = nil;
	short	oldVRefNum;
	long	oldDirID;
	char	filename[64];

	error = SetDefault(spec->vRefNum, spec->parID, &oldVRefNum, &oldDirID);
	if ( error == noErr )
	{
		Pstrcpy2C (filename, spec->name, 63 );
		fp = fopen( filename, mode);
		error = RestoreDefault(oldVRefNum, oldDirID);
	}
	
	return fp;
}


//----------------------------------------------------------------------
//
//	RewindFile
//
//----------------------------------------------------------------------
OSErr RewindFile ( FILE *fp )
{
	return (fseek(fp, 0, SEEK_SET) ? eofErr : noErr);
}


//----------------------------------------------------------------------
//
//	FileLength
//
//----------------------------------------------------------------------
OSErr FileLength ( FILE *fp, long *len )
{
	OSErr	theErr=noErr;
	long	old = -1;
	
	old = ftell(fp);
	require_action(old!=-1, bail, theErr=eofErr; );
	
	theErr = fseek(fp, 0, SEEK_END) ? eofErr : noErr;
	require(theErr==noErr, bail);
	
	*len = ftell(fp);
	require_action(old!=-1, bail, theErr=eofErr; );
	
bail:
	
	if (old!=-1)
		(void) fseek(fp,old,SEEK_SET);
	
	return theErr;
}


#pragma mark -
//----------------------------------------------------------------------
//
//	ReadSize -
//
//----------------------------------------------------------------------
OSErr ReadSize ( FILE *fp, UInt32 size, Ptr data )
{
	OSErr	theErr=noErr;
	UInt32	newsize;
	
	newsize = fread(data,1,size,fp);
	if (ferror(fp))
		theErr = readErr;
	else if (feof(fp) || newsize!=size)
		theErr = eofErr;
	return theErr;
}

//----------------------------------------------------------------------
//
//	WriteSize
//
//----------------------------------------------------------------------
OSErr WriteSize ( FILE *fp, UInt32 size, Ptr data )
{
	OSErr	theErr=noErr;
	UInt32	newsize;
	
	newsize = fwrite(data,1,size,fp);
	if (ferror(fp))
		theErr = writErr;
	else if (feof(fp) || newsize!=size)
		theErr = eofErr;
	return theErr;
}

//----------------------------------------------------------------------
//
//	SkipSize
//
//----------------------------------------------------------------------
OSErr SkipSize ( FILE *fp, UInt32 size )
{
	return (fseek(fp, size, SEEK_CUR) ? eofErr : noErr);
}

//----------------------------------------------------------------------
//
//	CopySize
//
//----------------------------------------------------------------------
OSErr CopySize ( FILE *from, FILE *into, UInt32 size, Ptr data )
{
	OSErr		theErr = noErr;
	Ptr			theData = nil;
	
	if (from && into && !data)
	{
		theData = SmartNewPtr(size, &theErr);
		require(theErr==noErr, bail);
		data = theData;
	}
		
	if (from && data)
	{
		theErr = ReadSize(from, size, data);
		require(theErr==noErr, bail);
	}
	if (from && !data)
	{
		theErr = SkipSize(from, size);
		require(theErr==noErr, bail);
	}
	if (into && data)
	{
		theErr = WriteSize(into, size, data);
		require(theErr==noErr, bail);
	}
	
bail:
	
	SmartDisposePtr(&theData);
	
	return theErr;
}


#pragma mark -
//----------------------------------------------------------------------
//
//	ReadByte
//
//----------------------------------------------------------------------
OSErr ReadByte ( FILE *fp, UInt8 *data )
{
	OSErr		theErr=noErr;
	UInt8		theData=0;
	
	theErr = ReadSize(fp, sizeof(theData), (Ptr)&theData);
	if (data) *data = theData;
	return theErr;
}

//----------------------------------------------------------------------
//
//	WriteByte
//
//----------------------------------------------------------------------
OSErr WriteByte ( FILE *fp, UInt8 data )
{
	return WriteSize(fp, sizeof(UInt8), (Ptr)&data);
}

//----------------------------------------------------------------------
//
//	CopyByte
//
//----------------------------------------------------------------------
OSErr CopyByte ( FILE *from, FILE *into, UInt8 *data)
{
	UInt8		theData=0;
	
	if (from && into && !data)
		data = &theData;
	return CopySize(from, into, sizeof(UInt8), (Ptr)data );
}


#pragma mark -
//----------------------------------------------------------------------
//
//	ReadWord
//
//----------------------------------------------------------------------
OSErr ReadWord ( FILE *fp, UInt16 *data, int endian )
{
	OSErr		theErr=noErr;
	UInt16		theData=0;

	theErr = ReadSize(fp, sizeof(UInt16), (Ptr)&theData);
	
	switch (endian)
	{
		case kEndianN:
			break;
		case kEndianB:
			theData = EndianU16_BtoN(theData);
			break;
		case kEndianL:
			theData = EndianU16_LtoN(theData);
			break;
		default:
			check(false);
	}
	
	if (data) *data = theData;
	return theErr;
}

//----------------------------------------------------------------------
//
//	WriteWord
//
//----------------------------------------------------------------------
OSErr WriteWord ( FILE *fp, UInt16 data, int endian )
{
	switch (endian)
	{
		case kEndianN:
			break;
		case kEndianB:
			data = EndianU16_NtoB(data);
			break;
		case kEndianL:
			data = EndianU16_NtoL(data);
			break;
		default:
			check(false);
	}
	
	return WriteSize(fp, sizeof(UInt16), (Ptr)&data);
}

//----------------------------------------------------------------------
//
//	CopyWord
//
//----------------------------------------------------------------------
OSErr CopyWord ( FILE *from, FILE *into, UInt16 *data, int endian )
{
	OSErr		theErr = noErr;
	UInt16		theData;
	
	if (from && into && !data)
		data = &theData;
		
	if (from && data)
	{
		theErr = ReadWord(from, data, endian);
		require(theErr==noErr, bail);
	}
	if (from && !data)
	{
		theErr = SkipSize(from, sizeof(UInt16));
		require(theErr==noErr, bail);
	}
	if (into && data)
	{
		theErr = WriteWord(into, *data, endian);
		require(theErr==noErr, bail);
	}
	
bail:
	
	return theErr;
}


#pragma mark -
//----------------------------------------------------------------------
//
//	ReadLong
//
//----------------------------------------------------------------------
OSErr ReadLong ( FILE *fp, UInt32 *data, int endian )
{
	OSErr		theErr=noErr;
	UInt32		theData=0;
	
	theErr = ReadSize(fp, sizeof(UInt32), (Ptr)&theData);
	
	switch (endian)
	{
		case kEndianN:
			break;
		case kEndianB:
			theData = EndianU32_BtoN(theData);
			break;
		case kEndianL:
			theData = EndianU32_LtoN(theData);
			break;
		default:
			check(false);
	}
	
	if (data) *data = theData;
	return theErr;
}

//----------------------------------------------------------------------
//
//	WriteLong
//
//----------------------------------------------------------------------
OSErr WriteLong ( FILE *fp, UInt32 data, int endian )
{
	switch (endian)
	{
		case kEndianN:
			break;
		case kEndianB:
			data = EndianU32_NtoB(data);
			break;
		case kEndianL:
			data = EndianU32_NtoL(data);
			break;
		default:
			check(false);
	}
	
	return WriteSize(fp, sizeof(UInt32), (Ptr)&data);
}

//----------------------------------------------------------------------
//
//	CopyLong
//
//----------------------------------------------------------------------
OSErr CopyLong ( FILE *from, FILE *into, UInt32 *data, int endian )
{
	OSErr		theErr = noErr;
	UInt32		theData;
	
	if (from && into && !data)
		data = &theData;
		
	if (from && data)
	{
		theErr = ReadLong(from, data, endian);
		require(theErr==noErr, bail);
	}
	if (from && !data)
	{
		theErr = SkipSize(from, sizeof(UInt32));
		require(theErr==noErr, bail);
	}
	if (into && data)
	{
		theErr = WriteLong(into, *data, endian);
		require(theErr==noErr, bail);
	}
	
bail:
	
	return theErr;
}
