
#include <MacTypes.h>
#include <CMApplication.h>
#include <CMICCProfile.h>
#include <Aliases.h>
#include <Devices.h>			// For WCTabHandle definition
#include <Displays.h>			// For definition of AVIDType
#include <Errors.h>
#include <Folders.h>
#include <Gestalt.h>
#include <QuickDraw.h>
#include <TextUtils.h>

#include <fp.h>
#include <stddef.h>

#include "GammaUtils.h"
#include "DebugAssert.h"



#if PRAGMA_STRUCT_ALIGN
	#pragma options align=mac68k
#elif PRAGMA_STRUCT_PACKPUSH
	#pragma pack(push, 2)
#elif PRAGMA_STRUCT_PACK
	#pragma pack(2)
#endif



			//	August 15, 1990 jmp mod -- added this definition for ease of use with PBStatus,
			//	which is replacing the calls to WeirdStatus (i.e. theres nothing weird
			//	about _Status except in the high-level glue, and who needs it anyway).
typedef struct VideoParamBlock
{
	QElemPtr	qLink;
	SInt16 		qType;
	SInt16 		ioTrap;
	Ptr			ioCmdAddr;
	ProcPtr		ioCompletion;
	OSErr		ioResult;
	StringPtr	ioNamePtr;
	SInt16 		ioVRefNum;
	SInt16 		ioRefNum;
	SInt16 		csCode;
	Ptr			csParam;
} VideoParamBlock, *VideoParamBlockPtr;


#if PRAGMA_STRUCT_ALIGN
	#pragma options align=reset
#elif PRAGMA_STRUCT_PACKPUSH
	#pragma pack(pop)
#elif PRAGMA_STRUCT_PACK
	#pragma pack()
#endif


#define kIntRoundOff			0.5
#define kFullGammaCnt			0x100


//---------------------------------------------------------------------
//	Static function prototypes
//---------------------------------------------------------------------

static OSErr ConvertSystemGammaToTable	(	CMVideoCardGammaTable*	table,
											GammaTblPtr				sysGamma,
											UInt32*					size );
static OSErr FillSystemGammaFormula		(	CMVideoCardGammaFormula* formula,
											GammaTblPtr				sysGamma);
static OSErr FillSystemGammaTable		(	CMVideoCardGammaTable*	table,
											GammaTblPtr				sysGamma);
static OSErr DoDMGetOldGammaByAVID		(	AVIDType				theAVID,
											GammaTblPtr*			gamma );
static OSErr DoDMGetBestOldGammaByAVID	(	AVIDType				theAVID,
											GammaTblPtr*			gamma );
static OSErr DoDMSetOldGammaByAVID		(	AVIDType				theAVID,
											GammaTblPtr				gamma );

static int memdif(const void * src1, const void * src2, size_t n);

#pragma mark -

//---------------------------------------------------------------------
//	Functions
//---------------------------------------------------------------------

//--------------------------------------------------------------------- DoCMGetGammaByAVID

OSErr DoCMGetGammaByAVID (AVIDType theAVID, CMVideoCardGamma *gamma, UInt32 *size)
{
	OSErr			theErr;
	GammaTblPtr		sysGamma;
	UInt32			tableSize;
	
	theErr = noErr;
	sysGamma = nil;
	
				// Check parameters
	require_action(theAVID && size, bail, theErr = paramErr;);
	
				// Get the current old-style gamma
	theErr = DoDMGetOldGammaByAVID(theAVID, &sysGamma);
	require(theErr == noErr, bail);
	
	if (gamma)
	{
		CMVideoCardGammaTable	*gammaTable;
		
		gamma->tagType = cmVideoCardGammaTableType;
		gammaTable = &(gamma->u.table);
		tableSize = (*size) - 4;	// ofsetof(CMVideoCardGamma,u) = 4
		theErr = ConvertSystemGammaToTable(gammaTable, sysGamma, &tableSize);
	}
	else
	{
		theErr = ConvertSystemGammaToTable(nil, sysGamma, &tableSize);
		*size = tableSize + 4;	// ofsetof(CMVideoCardGamma,u) = 4
	}
	
	if (theErr != noErr)
		goto bail;
	
bail:
	
	if (sysGamma) DisposePtr((Ptr) sysGamma);
	
	return theErr;
}


//--------------------------------------------------------------------- DoCMSetGammaByAVID

OSErr DoCMSetGammaByAVID (AVIDType theAVID, CMVideoCardGamma *gamma)
{
	OSErr			theErr;
	GammaTblPtr		sysGamma;
	
	theErr = noErr;
	sysGamma = nil;
	
				// Check parameters
	require_action(theAVID && gamma, bail, theErr = paramErr;);
	
				// Get the best old-style gamma.
	theErr = DoDMGetBestOldGammaByAVID(theAVID, &sysGamma);
	require(theErr == noErr, bail);
	
				// Convert to the old-style gamma.
	if (gamma->tagType == cmVideoCardGammaFormulaType)
	{
		theErr = FillSystemGammaFormula(&(gamma->u.formula), sysGamma);
	}
	else if (gamma->tagType == cmVideoCardGammaTableType)
	{
		theErr = FillSystemGammaTable(&(gamma->u.table), sysGamma);
	}
	else
	{
		theErr = paramErr;
		check(false);
	}
	
	if (theErr != noErr)
		goto bail;
	
				// Set the old-style gamma
	theErr = DoDMSetOldGammaByAVID(theAVID, sysGamma);
	
bail:
	
	if (sysGamma) DisposePtr((Ptr) sysGamma);
	
	return theErr;
}


//--------------------------------------------------------------------- ConvertSystemGammaToTable

static OSErr ConvertSystemGammaToTable (CMVideoCardGammaTable *table, GammaTblPtr sysGamma, UInt32 *size)
{
	CMVideoCardGammaTable	tableHeader;
	UInt32					theSize, theCount;
	OSErr					theErr;
	
	theErr = noErr;
	
				// Check parameters
	require_action(sysGamma && size, bail, theErr = paramErr;);
	require_action(sysGamma->gChanCnt == 1 || sysGamma->gChanCnt == 3, bail, theErr = paramErr;);
	require_action(sysGamma->gDataWidth >= 8 && sysGamma->gDataWidth <= 16, bail, theErr = paramErr;);
	require_action(sysGamma->gDataCnt >= 2, bail, theErr = paramErr;);
	
				// Determine number of channels. We may be able to reduce from 3 to 1
	tableHeader.channels = sysGamma->gChanCnt;
	if (sysGamma->gChanCnt == 3)
	{
		UInt32		compareSize;
		UInt8		*ptrA, *ptrB, *ptrC;
		
		if (sysGamma->gDataWidth >= 8)		// byte-packed gamma (> one byte per entry)
		{
			compareSize = (sysGamma->gDataWidth * sysGamma->gDataCnt) / 8;
			
			ptrA = ((UInt8*)sysGamma->gFormulaData) + sysGamma->gFormulaSize;
			ptrB = ptrA + compareSize;
			ptrC = ptrB + compareSize;
			
			if (memdif(ptrA,ptrB,compareSize)==0 &&
				memdif(ptrA,ptrC,compareSize)==0)
				tableHeader.channels = 1;	// We can reduce from 3 to 1
		}
	}
	
				// Determine entries per channel
	tableHeader.entryCount = sysGamma->gDataCnt;
	
				// Determine bytes per entry
	if (sysGamma->gDataWidth <= 8)
		tableHeader.entrySize = 1;
	else
		tableHeader.entrySize = 2;
	
				// Calculate what the size should be
	theCount = tableHeader.channels * tableHeader.entryCount;
	theSize = (theCount * tableHeader.entrySize) + (offsetof(CMVideoCardGammaTable,data));
	
	if (table)
	{
				// Make sure we were given a big enough pointer.
		require_action(*size >= theSize, bail, theErr = paramErr;);
		
				// Copy the header over.
		*table = tableHeader;
		
				// Copy the data over.
		if (sysGamma->gDataWidth == 8)
		{
			UInt8	*srcPtr, *dstPtr; 
			
			srcPtr = ((UInt8 *)&(sysGamma->gFormulaData));
			dstPtr = ((UInt8 *)&(table->data));
			while (theCount--)
			{
				*dstPtr = *srcPtr;
				srcPtr++;
				dstPtr++;
			}
		}
		else if (sysGamma->gDataWidth == 16)
		{
			UInt16	*srcPtr, *dstPtr;
			
			srcPtr = ((UInt16 *)&(sysGamma->gFormulaData));
			dstPtr = ((UInt16 *)&(table->data));
			while (theCount--)
			{
				*dstPtr = *srcPtr;
				srcPtr++;
				dstPtr++;
			}
		}
		else // gDataWidth<16 && gDataWidth>8
		{
			UInt16	*srcPtr, *dstPtr, shiftH, shiftL; 
			srcPtr = ((UInt16*)&(sysGamma->gFormulaData));
			dstPtr = ((UInt16*)&(table->data));
			shiftH = 16 - (sysGamma->gDataWidth);
			shiftL = (sysGamma->gDataWidth) - shiftH;
			while (theCount--)
			{
				*dstPtr = ((*srcPtr)<<shiftH) | ((*srcPtr)>>shiftL);
				srcPtr++;
				dstPtr++;
			}
		}
	}
	else
		*size = theSize;
	
bail:
		
	return theErr;
}


//--------------------------------------------------------------------- FillSystemGammaTable

static OSErr FillSystemGammaTable (CMVideoCardGammaTable *table, GammaTblPtr sysGamma)
{
	OSErr			theErr;
	UInt16			index;
	UInt16			chan;
	UInt16		 	tblEntrySize;
	double			scale;
	double			tblScale;
	SInt16			tblChan;
	double			tblIndex;
	UInt32			tblIndexLo;
	UInt32			tblIndexHi;
	double			tblWeightLo;
	double			tblWeightHi;
	double			tblValue;
	UInt32			tblValueLo;
	UInt32			tblValueHi;
	char			*tblValuePtr;
	
	theErr = noErr;
	
				// Check parameters.
	require_action(sysGamma->gChanCnt == 1 || sysGamma->gChanCnt == 3, bail, theErr = paramErr;);
	require_action(sysGamma->gDataWidth >= 8 && sysGamma->gDataWidth <= 16, bail, theErr = paramErr;);
	require_action(sysGamma->gDataCnt >= 2, bail, theErr = paramErr;);
	
	require_action(table->channels == 1 || table->channels == 3, bail, theErr = paramErr;);
	require_action(table->entryCount >= 2, bail, theErr = paramErr;);
	
				// Compute scale of entries from the size of each entry in the table.
				// So, a gDataWidth of 8 (bits) can take on values from 0 to ((1 << 8) - 1) 255
	scale = (double)((1 << sysGamma->gDataWidth) - 1);
	
				// The CMVideoCardGammaTable uses entrySize for the number of bytes for each entry.
				// Only one or two bytes (8 or 16 bit) is currently supported.
	tblEntrySize = table->entrySize;
	require_action(tblEntrySize == 1 || tblEntrySize == 2, bail, theErr = paramErr;);
	
				// Like the variable 'scale' above, calculate the scale dictated by the table.
	tblScale = (double)((1 << (tblEntrySize * 8)) - 1);
	
				// Now we're going to walk through each of the system gammas entries.
				// Typically, this is 256 entries.
	for (index = 0; index < sysGamma->gDataCnt; index++)
	{
		UInt8		*val8Ptr;
		UInt16		*val16Ptr;
		
				// Two pointers are set up - one a byte pointer, the other a word pointer.
		val8Ptr = ((UInt8 *)&(sysGamma->gFormulaData)) + index;
		val16Ptr = ((UInt16 *)&(sysGamma->gFormulaData)) + index;
		
				// Handle the case here where there are a different number of entries in 
				// the system gamma versus the table.  (We'll have to do some scaling.)
		if (table->entryCount != sysGamma->gDataCnt)
		{
				// Get the ratio of table entries to system gamma entries.
			tblIndex = ((double)index);
			tblIndex *= ((double)(table->entryCount - 1));
			tblIndex /= ((double)(sysGamma->gDataCnt - 1));
			
			tblIndexLo = tblIndex;
			tblIndexHi = tblIndexLo + 1;
			if (tblIndexHi >= table->entryCount)
				tblIndexHi = (table->entryCount) - 1;
			
			tblWeightLo = tblIndex - (double)tblIndexLo;
			if (tblWeightLo < 0.0)
				tblWeightLo = 0.0;
			if (tblWeightLo > 1.0)
				tblWeightLo = 1.0;
			tblWeightHi = 1.0 - tblWeightLo;
		}
		
				// For each channel in the system gamma (usually red, green & blue)
		for (chan = 0; chan < sysGamma->gChanCnt; chan++)
		{
				// Handle case where system gamma has more channels than the table.
			if (chan >= table->channels)
				tblChan = table->channels - 1;
			else
				tblChan = chan;
			
				// If both the table and the system gamma have the same number of 
				// entries (i.e. no need to scale - probably both have 256 entries).
			if (table->entryCount == sysGamma->gDataCnt)
			{
				tblValuePtr = table->data + (tblEntrySize * (tblChan * (table->entryCount) + index));
				
				// Determine whether to use 8 bit or 16 bit pointer.
				if (tblEntrySize == 1)
					tblValue = *((UInt8 *)tblValuePtr);
				else
					tblValue = *((UInt16 *)tblValuePtr);
				
				tblValue /= tblScale;
			}
			else	// Handle the tricky case of scaling from one number of entries to another.
			{
				tblValuePtr = table->data + (tblEntrySize*(tblChan*(table->entryCount)+tblIndexLo));
				tblValueLo = (tblEntrySize == 1)?(*((UInt8*)tblValuePtr)):(*((UInt16*)tblValuePtr));
				
				tblValuePtr = table->data + (tblEntrySize*(tblChan*(table->entryCount)+tblIndexHi));
				tblValueHi = (tblEntrySize == 1)?(*((UInt8*)tblValuePtr)):(*((UInt16*)tblValuePtr));
				
				tblValue = (tblWeightHi) * ((double)tblValueLo);
				tblValue += (tblWeightLo) * ((double)tblValueHi);
				tblValue /= tblScale;
			}
			
			if (sysGamma->gDataWidth == 8)
			{
				*val8Ptr = (UInt8)((tblValue * scale) + kIntRoundOff);
				val8Ptr += (sysGamma->gDataCnt);
			}
			else // sysGamma->gDataWidth <= 16
			{
				*val16Ptr = (UInt16)((tblValue * scale) + kIntRoundOff);
				val16Ptr += (sysGamma->gDataCnt);
			}
		}
	}
	
bail:
	
	return theErr;
}


//--------------------------------------------------------------------- FillSystemGammaFormula

static OSErr FillSystemGammaFormula (CMVideoCardGammaFormula *formula, GammaTblPtr sysGamma)
{
	OSErr		theErr;
	UInt16		index;
	UInt16		chan;
	double		ind;
	double		val;
	double		minR, maxR, gamR;
	double		minG, maxG, gamG;
	double		minB, maxB, gamB;
	double		scale;
	
	require_action(sysGamma->gChanCnt == 1 || sysGamma->gChanCnt == 3, bail, theErr = paramErr;);
	require_action(sysGamma->gDataWidth >= 8 && sysGamma->gDataWidth <= 16, bail, theErr = paramErr;);
	require_action(sysGamma->gDataCnt >= 2, bail, theErr = paramErr;);
	
	theErr = noErr;
	scale = (double)((1 << sysGamma->gDataWidth) - 1);
	
	minR = ((double)formula->redMin) / 65536.0;
	maxR = ((double)formula->redMax) / 65536.0;
	gamR = ((double)formula->redGamma) / 65536.0;
	minG = ((double)formula->greenMin) / 65536.0;
	maxG = ((double)formula->greenMax) / 65536.0;
	gamG = ((double)formula->greenGamma) / 65536.0;
	minB = ((double)formula->blueMin) / 65536.0;
	maxB = ((double)formula->blueMax) / 65536.0;
	gamB = ((double)formula->blueGamma) / 65536.0;
	
	if (sysGamma->gChanCnt == 1)
	{
		minR = (minR + minG + minB) / 3.0;
		maxR = (maxR + maxG + maxB) / 3.0;
		gamR = (gamR + gamG + gamB) / 3.0;
	}
	
	for (index = 0; index < sysGamma->gDataCnt; index++)
	{
		UInt8		*val8Ptr;
		UInt16		*val16Ptr;
		
		val8Ptr = ((UInt8 *)&(sysGamma->gFormulaData)) + index;
		val16Ptr = ((UInt16 *)&(sysGamma->gFormulaData)) + index;
		
		ind = ((double)index) / ((double)(sysGamma->gDataCnt - 1));
		
		for (chan = 0; chan < sysGamma->gChanCnt; chan++)
		{
			val = 0;
			
			if (chan == 0)
				val = minR + ((maxR - minR) * pow(ind, gamR));
			else if (chan == 1)
				val = minG + ((maxG - minG) * pow(ind, gamG));
			else if (chan == 2)
				val = minB + ((maxB - minB) * pow(ind, gamB));
			else
				val = 0.0;
			
			if (sysGamma->gDataWidth == 8)
			{
				val *= scale;
				val += kIntRoundOff;
				*val8Ptr = (UInt8)val;
				val8Ptr += (sysGamma->gDataCnt);
			}
			else // sysGamma->gDataWidth <= 16
			{
				*val16Ptr = (UInt16)(val * scale) + kIntRoundOff;
				val16Ptr += (sysGamma->gDataCnt);
			}
		}
	}
	
bail:
	
	return theErr;
}


//--------------------------------------------------------------------- DoDMGetOldGammaByAVID

static OSErr DoDMGetOldGammaByAVID (AVIDType theAVID, GammaTblPtr *gamma)
{
	VideoParamBlock		pBlock;
	VDGammaRecord		gammaRecord;
	GDHandle			gDevice;
	GammaTblPtr			gamTblPtr;
	UInt32				gammaSize;
	UInt32				dataSize;
	UInt32				dataCount;
	OSErr				theErr;
	SInt16				refNum;
	
	theErr = noErr;
	
				// Check parameters
	require_action(theAVID && gamma, bail, theErr = paramErr;);
	
	*gamma = nil;
	
				// Get the current gDevice and video driver refnum
	theErr = DMGetGDeviceByDisplayID(theAVID, &gDevice, false);
	require(theErr == noErr, bail);
	
	refNum = (**gDevice).gdRefNum;
	require_action(refNum, bail, theErr=paramErr;);
	
				// Send a request to get the gamma table.
	gammaRecord.csGTable = nil;
	pBlock.ioNamePtr = nil;
	pBlock.ioRefNum = refNum;
	pBlock.csCode = cscGetGamma;	
	pBlock.csParam = (Ptr)&gammaRecord;
	
	theErr = PBStatusSync((ParmBlkPtr)&pBlock);
	require(theErr == noErr, bail);
	
				// I've looked at some drivers (such as DartVSCDriver) and it is possible
				// for them to return noErr and return a nil ptr.
	if (nil == gammaRecord.csGTable)
		theErr = statusErr;
	require(theErr == noErr, bail);
	
	gamTblPtr = (GammaTblPtr)(gammaRecord.csGTable); 
	
				// Calculate size of gamma table.
				// First, figure out the number of total entries (channels * entries per channel).
	dataCount = gamTblPtr->gDataCnt * gamTblPtr->gChanCnt;
	
				// Next, using the size of each entry, calculate the total number of bytes
				// that all the entries (dataCount) require.
	if (gamTblPtr->gDataWidth < 8)
		dataSize = ((gamTblPtr->gDataWidth * dataCount) + 7) / 8;	// bit-packed gamma (< one byte per entry)
	else if (gamTblPtr->gDataWidth == 8)
		dataSize = dataCount;										// simple gamma (one byte per entry)
	else
		dataSize = ((gamTblPtr->gDataWidth + 7) / 8) * dataCount;	// byte-packed gamma (> one byte per entry)
	
				// Calculate the total number of bytes of the works.
	gammaSize = dataSize + gamTblPtr->gFormulaSize + offsetof(GammaTbl,gFormulaData);
	
	*gamma = (GammaTblPtr)NewPtrClear(gammaSize);			// Caller will dispose of this ptr
	require_action(*gamma, bail, theErr=MemError(); );
	
	BlockMoveData(gamTblPtr, *gamma, gammaSize);
	
bail:
	
	return theErr;
}


//--------------------------------------------------------------------- DoDMGetBestOldGammaByAVID

static OSErr DoDMGetBestOldGammaByAVID (AVIDType theAVID, GammaTblPtr *gamma)
{
				// For now this just gets the current old-style gamma
	return DoDMGetOldGammaByAVID(theAVID, gamma);
}


//--------------------------------------------------------------------- DoDMSetOldGammaByAVID

static OSErr DoDMSetOldGammaByAVID (AVIDType theAVID, GammaTblPtr gamma)
{
	OSErr				theErr;
	VDGammaRecord		gammaRecord;
	GDHandle			gDevice;
	SInt16				refNum;
	Ptr					csParamPtr;
	
	theErr = noErr;
	
				// Check parameters
	require_action(theAVID && gamma, bail, theErr = paramErr;);
	
				// Get the current gDevice and video driver refnum
	theErr = DMGetGDeviceByDisplayID(theAVID, &gDevice, false);
	require(theErr == noErr, bail);
	
	refNum = (**gDevice).gdRefNum;
	require_action(refNum, bail, theErr = paramErr;);
	
				// Send a request to set the gamma table.
	gammaRecord.csGTable = (Ptr)gamma;
	csParamPtr = (Ptr)&gammaRecord;
	theErr = Control(refNum, cscSetGamma, &csParamPtr);
	require(theErr == noErr, bail);
	
				// Do SetEntries to make it happen on < 16 bit video.
	if ((**gDevice).gdType == clutType)
	{
		PixMapHandle		thePixMap;
		CTabHandle			theColorTable;
		VDSetEntryRecord	theSetEntryRec;
		
		thePixMap = (**gDevice).gdPMap;
		theColorTable = (**thePixMap).pmTable;
		theSetEntryRec.csTable = (ColorSpec*)(**theColorTable).ctTable;
		theSetEntryRec.csStart = 0;
		theSetEntryRec.csCount = (**theColorTable).ctSize;
		
		csParamPtr = (Ptr)&theSetEntryRec;
		theErr = Control(refNum, cscSetEntries, &csParamPtr);
		require(theErr == noErr, bail);
	}
	
				// After setting up the gamma, we need to call Display Manager in order
				// for the screen settings in effect at this moment to be saved somewhere.
				// Hope this actually works...
	theErr = DMSaveScreenPrefs(0, 0, 0);
	require(theErr == noErr, bail);
	
bail:
	
	return theErr;
}


static int memdif(const void * src1, const void * src2, size_t n)
{
	const	unsigned char * p1;
	const	unsigned char * p2;

#if !__POWERPC__
	
	for (p1 = (const unsigned char *) src1, p2 = (const unsigned char *) src2, n++; --n;)
		if (*p1++ != *p2++)
			return 1;
	
#else
	
	for (p1 = (const unsigned char *) src1 - 1, p2 = (const unsigned char *) src2 - 1, n++; --n;)
		if (*++p1 != *++p2)
			return 1;

#endif
	
	return(0);
}

